import json
import os
import shlex
import subprocess
import sys
import time
import uuid
from pathlib import Path
from subprocess import Popen
from typing import Callable, Dict, List, Optional, Union

import inquirer
from rich.console import Console
from ruamel import yaml
from ruamel.yaml import CommentedMap, CommentedSeq

from piperider_cli import raise_exception_when_directory_not_writable, round_trip_load_yaml, safe_load_yaml, dbtutil
from piperider_cli.datasource import DATASOURCE_PROVIDERS, DataSource
from piperider_cli.error import \
    PipeRiderConfigError, \
    PipeRiderConfigTypeError, \
    PipeRiderInvalidDataSourceError, \
    DbtProjectNotFoundError, \
    DbtProfileNotFoundError

# ref: https://docs.getdbt.com/dbt-cli/configure-your-profile
DBT_PROFILES_DIR_DEFAULT = '~/.dbt/'
DBT_PROFILE_FILE = 'profiles.yml'
PIPERIDER_CONFIG_FILE = 'config.yml'
PIPERIDER_CREDENTIALS_FILE = 'credentials.yml'


class _FileSystem:

    def __init__(self):
        self.working_directory = os.getcwd()
        self.dbt_profiles_dir = None

    def set_working_directory(self, working_directory: str):
        if os.path.exists(working_directory):
            self.working_directory = working_directory if os.path.isdir(working_directory) else os.path.dirname(
                working_directory)
        else:
            raise FileNotFoundError(f'the directory[{working_directory}] is not existing.')

    def set_dbt_profiles_dir(self, dbt_profiles_dir: str):
        if os.path.exists(os.path.join(dbt_profiles_dir, DBT_PROFILE_FILE)):
            self.dbt_profiles_dir = dbt_profiles_dir
        else:
            raise FileNotFoundError(f'[DBT Profiles Dir] the directory[{dbt_profiles_dir}] is not existing.')

    @property
    def WORKING_DIRECTORY(self):
        return self.working_directory

    @property
    def DBT_PROFILES_DIR(self):
        if self.dbt_profiles_dir:
            return self.dbt_profiles_dir
        elif os.getenv('DBT_PROFILES_DIR'):
            return os.getenv('DBT_PROFILES_DIR')
        elif os.path.exists(os.path.join(self.working_directory, DBT_PROFILE_FILE)):
            return self.working_directory
        elif os.path.exists(os.path.join(os.getcwd(), DBT_PROFILE_FILE)):
            return os.getcwd()
        else:
            return os.path.expanduser(DBT_PROFILES_DIR_DEFAULT)

    @property
    def PIPERIDER_WORKSPACE_NAME(self):
        return '.piperider'

    @property
    def PIPERIDER_WORKSPACE_PATH(self):
        return os.path.join(self.working_directory, self.PIPERIDER_WORKSPACE_NAME)

    @property
    def PIPERIDER_CONFIG_PATH(self):
        return os.path.join(self.PIPERIDER_WORKSPACE_PATH, PIPERIDER_CONFIG_FILE)

    @property
    def PIPERIDER_CREDENTIALS_PATH(self):
        return os.path.join(self.PIPERIDER_WORKSPACE_PATH, PIPERIDER_CREDENTIALS_FILE)

    @property
    def PIPERIDER_ASSERTION_SEARCH_PATH(self):
        return os.path.join(self.PIPERIDER_WORKSPACE_PATH, 'assertions')

    @property
    def PIPERIDER_ASSERTION_PLUGIN_PATH(self):
        return os.path.join(self.PIPERIDER_WORKSPACE_PATH, 'plugins')

    @property
    def PIPERIDER_RECIPES_PATH(self):
        return os.path.join(self.PIPERIDER_WORKSPACE_PATH, 'compare')

    @property
    def DEFAULT_RECIPE_PATH(self):
        return os.path.join(self.PIPERIDER_RECIPES_PATH, 'default.yml')

    @property
    def piperider_default_report_dir(self):
        return os.path.join(os.getcwd(), self.PIPERIDER_WORKSPACE_NAME)


FileSystem = _FileSystem()


class ReportDirectory:

    def __init__(self, config: "Configuration", **kwargs):
        self.report_dir = config.report_dir
        if kwargs.get('report_dir', None) is not None:
            self.report_dir = self.normalize_report_dir(kwargs.get('report_dir'))

        raise_exception_when_directory_not_writable(self.report_dir)

    @staticmethod
    def normalize_report_dir(dirname: str):
        """
        the "." always refer to `.piperider` not the current working directory
        """
        if dirname is None or dirname.strip() == '':
            dirname = '.'
        if dirname.startswith('.'):
            return os.path.abspath(
                os.path.join(FileSystem.WORKING_DIRECTORY, FileSystem.PIPERIDER_WORKSPACE_NAME, dirname))
        return os.path.abspath(dirname)

    def get_output_dir(self):
        if self.report_dir is None:
            return os.path.join(FileSystem.piperider_default_report_dir, 'outputs')
        return os.path.join(self.report_dir, 'outputs')

    def get_comparison_dir(self):
        if self.report_dir is None:
            return os.path.join(FileSystem.piperider_default_report_dir, 'comparisons')
        return os.path.join(self.report_dir, 'comparisons')

    def get_report_dir(self):
        return self.report_dir


def is_piperider_workspace_exist(workspace_path: str = None) -> bool:
    workspace_path = workspace_path or FileSystem.PIPERIDER_WORKSPACE_PATH
    if not os.path.exists(workspace_path):
        return False
    elif not os.path.exists(os.path.join(workspace_path, 'config.yml')):
        return False

    return True


class TelemetryIdResolver:

    def __init__(self):
        self.telemetry_id = None
        self.resolvers = [
            self.resolve_from_local_state,
            self.resolve_from_cloud_state,
            self.resolve_from_legacy_config,
            self.resolve_from_git_remote_url,
            self.resolve_by_new_id,
        ]
        self.show_debug_message = os.environ.get('SHOW_PIPERIDER_TELEMETRY_ID_RESOLVED_TIME') is not None

    def resolve(self, configuration: "Configuration") -> str:
        if self.telemetry_id is not None:
            return self.telemetry_id

        def _execute():
            for r in self.resolvers:
                callback: Callable[["Configuration"], str] = r
                telemetry_id = None
                try:
                    telemetry_id = callback(configuration)
                    if self.show_debug_message and telemetry_id is not None:
                        print(f"resolve telemetry-id by: {callback}")
                        print(f"telemetry-id: {telemetry_id}")
                except BaseException:
                    pass
                if telemetry_id is None:
                    continue

                self.telemetry_id = telemetry_id
                return self.telemetry_id

        t1 = time.time()
        try:
            return _execute()
        finally:
            t2 = time.time()
            if self.show_debug_message:
                print(f"resolve telemetry-id in {t2 - t1:.1} seconds")

    @staticmethod
    def resolve_from_local_state(configuration: "Configuration"):
        if configuration.report_directory_filesystem is None:
            return
        output_dir = configuration.report_directory_filesystem.get_output_dir()

        def _extract_id(run_json_file: str):
            with open(run_json_file) as fh:
                content: Dict = json.loads(fh.read())
                project_id = content.get('project_id')
                if project_id:
                    return project_id

        def _resolve_id_from_report_dir(directory):
            target_file = 'run.json'

            for root, dirs, files in os.walk(directory):
                if target_file in files:
                    file_path = os.path.join(root, target_file)
                    telemetry_id = _extract_id(file_path)
                    if telemetry_id:
                        return telemetry_id

        return _resolve_id_from_report_dir(output_dir)

    @staticmethod
    def resolve_from_cloud_state(configuration: "Configuration"):
        from piperider_cli.cloud_connector import CloudConnector, piperider_cloud
        if CloudConnector.is_login():
            default_project = piperider_cloud.get_default_project()
            datasource_id = piperider_cloud.get_datasource_id(default_project)
            if datasource_id:
                return datasource_id

    @staticmethod
    def resolve_from_legacy_config(configuration: "Configuration"):
        return configuration.telemetry_id

    @staticmethod
    def resolve_from_git_remote_url(configuration: "Configuration"):
        def _exec(command_line: str):
            cmd = shlex.split(command_line)
            proc = None
            try:
                proc = Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                outs, errs = proc.communicate()
            except BaseException as e:
                if proc:
                    proc.kill()
                    outs, errs = proc.communicate()
                else:
                    return None, e, 1

            if outs is not None:
                outs = outs.decode().strip()
            if errs is not None:
                errs = errs.decode().strip()
            return outs, errs, proc.returncode

        outs, errs, exit_code = _exec("git remote get-url origin")
        if exit_code != 0:
            return

        return uuid.uuid5(uuid.NAMESPACE_URL, outs).hex

    @staticmethod
    def resolve_by_new_id(configuration: "Configuration"):
        return uuid.uuid4().hex


configuration_instance: Optional["Configuration"] = None


class Configuration(object):
    """
    Configuration represents the config file in the piperider project
    at $PROJECT_ROOT/.piperider/config.yml
    """

    def __init__(self, dataSources: List[DataSource], **kwargs):
        self.dataSources: List[DataSource] = dataSources
        self.profiler_config = kwargs.get('profiler', {}) or {}
        self.includes = kwargs.get('includes', None)
        self.excludes = kwargs.get('excludes', None)
        self.include_views = kwargs.get('include_views', False)
        self.tables = kwargs.get('tables', {})

        # only the legacy project will set telemetry_id from config
        self.telemetry_id = kwargs.get('telemetry_id', None)
        self.report_dir = ReportDirectory.normalize_report_dir(kwargs.get('report_dir', '.'))
        self.report_directory_filesystem: Optional[ReportDirectory] = None

        self._verify_input_config()
        self.includes = [str(t) for t in self.includes] if self.includes else self.includes
        self.excludes = [str(t) for t in self.excludes] if self.excludes else self.excludes

        # global dbt config
        self.dbt = kwargs.get('dbt', None)
        self.telemetry_resolver = TelemetryIdResolver()

        # local default project config
        self.cloud_config = kwargs.get('cloud_config', None)

    def _verify_input_config(self):
        if self.profiler_config:
            limit = self.profiler_config.get('table', {}).get('limit', 0)
            if not isinstance(limit, int):
                raise PipeRiderConfigTypeError("profiler 'limit' should be an integer")

            duplicate_rows = self.profiler_config.get('table', {}).get('duplicateRows', False)
            if not isinstance(duplicate_rows, bool):
                raise PipeRiderConfigTypeError("profiler 'duplicateRows' should be an boolean")

        if self.includes is not None:
            if not isinstance(self.includes, List):
                raise PipeRiderConfigTypeError("'includes' should be a list of tables' name")

        if self.excludes is not None:
            if not isinstance(self.excludes, List):
                raise PipeRiderConfigTypeError("'excludes' should be a list of tables' name")

    def get_telemetry_id(self):
        if self.telemetry_id is not None:
            return self.telemetry_id

        self.telemetry_id = self.telemetry_resolver.resolve(self)
        assert self.telemetry_id is not None
        return self.telemetry_id

    def activate_report_directory(self, report_dir: str = None) -> ReportDirectory:
        if self.report_directory_filesystem is not None:
            return self.report_directory_filesystem

        fs = ReportDirectory(self, report_dir=report_dir)
        self.report_directory_filesystem = fs
        return self.report_directory_filesystem

    @staticmethod
    def update_config(key: str, update_values: Union[dict, str]):
        _yml = yaml.YAML()

        with open(FileSystem.PIPERIDER_CONFIG_PATH, 'r') as f:
            config = _yml.load(f)

        config[key] = update_values
        with open(FileSystem.PIPERIDER_CONFIG_PATH, 'w+') as f:
            _yml.dump(config, f)

    @classmethod
    def from_dbt_project(cls, dbt_project_path, dbt_profiles_dir=None):
        """
        build configuration from the existing dbt project

        :param dbt_project_path:
        :param dbt_profiles_dir:
        :return:
        """

        # Precedence reference
        # https://docs.getdbt.com/docs/get-started/connection-profiles#advanced-customizing-a-profile-directory
        if dbt_profiles_dir:
            profile_dir = dbt_profiles_dir
        else:
            profile_dir = FileSystem.DBT_PROFILES_DIR

        dbt_profile_path = os.path.join(profile_dir, DBT_PROFILE_FILE)

        if not os.path.exists(dbt_project_path):
            raise DbtProjectNotFoundError(dbt_project_path)

        from piperider_cli.dbtutil import load_dbt_project
        dbt_project = load_dbt_project(dbt_project_path)

        if not os.path.exists(os.path.expanduser(dbt_profile_path)):
            raise DbtProfileNotFoundError(dbt_profile_path)

        dbt_profile = dbtutil.load_dbt_profile(os.path.expanduser(dbt_profile_path))

        console = Console()
        profile_name = dbt_project.get('profile', '')
        profile_content = dbt_profile.get(profile_name, None)

        if profile_content is None:
            console.print("[bold red]Error:[/bold red] "
                          f"Could not find profile named '{profile_name}'")
            sys.exit(1)
        target_name = profile_content.get('target', 'default')
        if target_name not in list(dbt_profile.get(profile_name, {}).get('outputs', {}).keys()):
            console.print("[bold red]Error:[/bold red] "
                          f"The profile '{profile_name}' does not have a target named '{target_name}'.\n"
                          "Please check the dbt profile format.")
            sys.exit(1)
        credential = dbtutil.load_credential_from_dbt_profile(dbt_profile, profile_name, target_name)
        type_name = credential.get('type')
        dbt = {
            'projectDir': os.path.relpath(os.path.dirname(dbt_project_path), FileSystem.WORKING_DIRECTORY),
            'tag': 'piperider',
        }

        if dbt_profiles_dir:
            dbt['profilesDir'] = dbt_profiles_dir

        if type_name not in DATASOURCE_PROVIDERS:
            raise PipeRiderInvalidDataSourceError(type_name, dbt_profile_path)

        # Set 'pass' as the alias of 'password'
        if credential.get('pass') and credential.get('password') is None:
            credential['password'] = credential.pop('pass')

        datasource_class = DATASOURCE_PROVIDERS[type_name]
        datasource: DataSource = datasource_class(name=profile_name, dbt=dbt, credential=credential)
        datasource.show_installation_information()

        return cls(dataSources=[datasource])

    @classmethod
    def instance(cls, piperider_config_path=None):
        piperider_working_directory = cls.search_piperider_project_path()
        if piperider_working_directory:
            FileSystem.set_working_directory(piperider_working_directory)
        piperider_config_path = piperider_config_path or FileSystem.PIPERIDER_CONFIG_PATH
        global configuration_instance
        if configuration_instance is not None:
            return configuration_instance
        configuration_instance = cls._load(piperider_config_path)
        return configuration_instance

    @classmethod
    def search_piperider_project_path(cls) -> str:
        paths = list(Path.cwd().parents)
        paths.insert(0, Path.cwd())
        return next(
            (str(x) for x in paths if (x / FileSystem.PIPERIDER_WORKSPACE_NAME / PIPERIDER_CONFIG_FILE).exists()),
            None)

    @classmethod
    def _load(cls, piperider_config_path=None):
        """
        load from the existing configuration

        :return:
        """
        credentials = None
        piperider_config_path = piperider_config_path or FileSystem.PIPERIDER_CONFIG_PATH

        config = safe_load_yaml(piperider_config_path)
        if config is None:
            raise PipeRiderConfigError(piperider_config_path)

        data_sources: List[DataSource] = []
        ds_basic_config = ['name', 'type', 'dbt']
        for ds in config.get('dataSources', []):
            type_name = ds.get('type')
            if type_name not in DATASOURCE_PROVIDERS:
                raise PipeRiderInvalidDataSourceError(type_name, piperider_config_path)

            datasource_class = DATASOURCE_PROVIDERS[type_name]
            credential = {}
            for config_key in ds:
                if config_key in ds_basic_config:
                    continue
                credential[config_key] = ds[config_key]

            dbt = ds.get('dbt')
            if dbt:
                # Precedence reference
                # https://docs.getdbt.com/docs/get-started/connection-profiles#advanced-customizing-a-profile-directory
                if dbt.get('profilesDir'):
                    profile_dir = dbt.get('profilesDir')
                else:
                    profile_dir = FileSystem.DBT_PROFILES_DIR
                profile_path = os.path.join(profile_dir, DBT_PROFILE_FILE)
                if '~' in profile_path:
                    profile_path = os.path.expanduser(profile_path)
                profile = dbtutil.load_dbt_profile(profile_path)
                profile_name = dbt.get('profile')
                target_name = dbt.get('target')
                credential.update(dbtutil.load_credential_from_dbt_profile(profile, profile_name, target_name))
                # TODO: extract duplicate code from func 'from_dbt_project'
                if credential.get('pass') and credential.get('password') is None:
                    credential['password'] = credential.pop('pass')
                data_source = datasource_class(name=ds.get('name'), dbt=dbt, credential=credential)
            else:
                try:
                    with open(FileSystem.PIPERIDER_CREDENTIALS_PATH, 'r') as fd:
                        credentials = yaml.safe_load(fd)
                        credential.update(credentials.get(ds.get('name'), {}))
                except FileNotFoundError:
                    pass
                except Exception:
                    raise
                data_source = datasource_class(name=ds.get('name'), credential=credential)
            data_sources.append(data_source)

        # load global dbt config
        dbt = config.get('dbt')
        if dbt:
            project_dir = config.get('dbt').get('projectDir')
            project = dbtutil.load_dbt_project(project_dir)
            profile_name = project.get('profile')

            # Precedence reference
            # https://docs.getdbt.com/docs/get-started/connection-profiles#advanced-customizing-a-profile-directory
            if dbt.get('profilesDir'):
                profile_dir = dbt.get('profilesDir')
            else:
                profile_dir = FileSystem.DBT_PROFILES_DIR

            profile_path = os.path.join(profile_dir, DBT_PROFILE_FILE)
            if '~' in profile_path:
                profile_path = os.path.expanduser(profile_path)
            profile = dbtutil.load_dbt_profile(profile_path)
            if profile.get(profile_name):
                target_names = list(profile.get(profile_name).get('outputs').keys())
                for target in target_names:
                    credential = dbtutil.load_credential_from_dbt_profile(profile, profile_name, target)
                    if credential.get('pass') and credential.get('password') is None:
                        credential['password'] = credential.pop('pass')
                    datasource_class = DATASOURCE_PROVIDERS[credential.get('type')]
                    data_source = datasource_class(
                        name=target,
                        dbt=dict(**dbt, profile=profile_name, target=target),
                        credential=credential
                    )
                    data_sources.append(data_source)
                # dbt behavior: dbt uses 'default' as target name if no target given in profiles.yml
                dbt['target'] = profile.get(profile_name).get('target', 'default')
                if dbt['target'] not in target_names:
                    console = Console()
                    console.print("[bold red]Error:[/bold red] "
                                  f"The profile '{profile_name}' does not have a target named '{dbt['target']}'.\n"
                                  "Please check the dbt profile format.")
                    sys.exit(1)

        return cls(
            dataSources=data_sources,
            profiler=config.get('profiler', {}),
            tables=config.get('tables', {}),
            includes=config.get('includes', None),
            excludes=config.get('excludes', None),
            include_views=config.get('include_views', False),
            telemetry_id=config.get('telemetry', {}).get('id'),
            report_dir=config.get('report_dir', '.'),
            dbt=dbt,
            cloud_config=config.get('cloud_config', {})
        )

    def flush_datasource(self, path):
        """
        flush the configuration to the file

        :param path:
        :return:
        """

        config = round_trip_load_yaml(path)
        if config is None:
            raise PipeRiderConfigError(path)

        def _get_exist_datasource(d, config):
            for ds in config.get('dataSources', []):
                if ds.get('name') == d.name:
                    return ds
            return None

        def _generate_datasource_config(d):
            datasource = dict(name=d.name, type=d.type_name)
            if d.args.get('dbt'):
                # dbt project
                datasource['dbt'] = d.args.get('dbt')
            else:
                # non-dbt project
                if d.credential_source == 'config':
                    datasource.update(**d.credential)
            return CommentedMap(datasource)

        flush_data_sources = CommentedSeq()

        for ds in self.dataSources:
            exist_ds = _get_exist_datasource(ds, config)
            if exist_ds:
                flush_data_sources.append(exist_ds)
            else:
                new_ds = _generate_datasource_config(ds)
                flush_data_sources.append(new_ds)
        config['dataSources'] = flush_data_sources

        if config['dataSources'][-1].ca.items == {}:
            # Only append a new line if the last datasource has no comment
            config.yaml_set_comment_before_after_key('profiler', before='\n')

        with open(path, 'w') as fd:
            yaml.YAML().dump(config, fd)
        pass

    def dump(self, path):
        """
        dump the configuration to the given path
        :param path:
        :return:
        """

        config = dict(
            dataSources=[],
            dbt=None,
        )

        d = self.dataSources[0]
        datasource = dict(name=d.name, type=d.type_name)
        if d.args.get('dbt'):
            # dbt project
            config['dbt'] = CommentedMap(d.args.get('dbt'))
        else:
            # non-dbt project
            if d.credential_source == 'config':
                datasource.update(**d.credential)
            config.pop('dbt')
            config['dataSources'].append(datasource)

        if self.cloud_config:
            config['cloud_config'] = self.cloud_config

        config_yaml = CommentedMap(config)

        with open(path, 'w') as fd:
            yaml.YAML().dump(config_yaml, fd)

    def dump_credentials(self, path, after_init_config=False):
        """
        dump the credentials to the given path
        :param path:
        :param after_init_config: config file has been written
        :return:
        """
        creds = dict()
        for d in self.dataSources:
            if after_init_config and d.credential_source == 'config':
                continue
            if d.args.get('dbt'):
                continue
            creds[d.name] = dict(type=d.type_name, **d.credential)

        if creds:
            with open(path, 'w') as fd:
                yaml.round_trip_dump(creds, fd)

    def to_sqlalchemy_config(self, datasource_name):
        # TODO we will convert a data source to a sqlalchemy parameters
        raise NotImplementedError

    def ask_for_datasource(self):
        if len(self.dataSources) == 0:
            return None
        elif len(self.dataSources) == 1:
            return self.dataSources[0]
        else:
            questions = [
                inquirer.List('datasource',
                              message="Please select a datasource",
                              choices=[(f'{d.name:20} (type: {d.type_name})', d) for d in self.dataSources],
                              carousel=True)
            ]
            answers = inquirer.prompt(questions, raise_keyboard_interrupt=True)
            return answers['datasource']

    def delete_datasource(self, datasource):
        if datasource in self.dataSources:
            self.dataSources.remove(datasource)



import json
import os
from typing import List

import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from rich.progress import Progress, TextColumn, BarColumn, DownloadColumn, TimeElapsedColumn
from ruamel import yaml

from piperider_cli import __version__
from piperider_cli.configuration import Configuration, FileSystem
from piperider_cli.error import PipeRiderNoDefaultProjectError, CloudReportError, PipeRiderConfigError
from piperider_cli.event import load_user_profile, update_user_profile

PIPERIDER_CLOUD_SERVICE = 'https://cloud.piperider.io/'

SERVICE_ENV_API_KEY = 'PIPERIDER_API_TOKEN'
SERVICE_ENV_SERVICE_KEY = 'PIPERIDER_API_SERVICE'

yml = yaml.YAML()


class PipeRiderProject(object):
    def __init__(self, project: dict):
        self.id = project.get('id')
        self.name = project.get('name')
        self.workspace_name = project.get('workspace_name')


class CloudServiceHelper:

    def __init__(self):
        self.api_token: str = None
        self.api_service: str = None
        self.user_profile = load_user_profile()
        if self.user_profile is None:
            return
        self.load_cloud_api_config()

    @property
    def cloud_host(self):
        if 'localhost' in self.api_service:
            return 'http://localhost:3000'
        return self.api_service

    def load_cloud_api_config(self):
        # load from the configuration file first
        # overwrite them if found env vars
        self.api_service = os.environ.get(SERVICE_ENV_SERVICE_KEY,
                                          self.user_profile.get('api_service',
                                                                PIPERIDER_CLOUD_SERVICE))
        self.api_token = os.environ.get(SERVICE_ENV_API_KEY,
                                        self.user_profile.get('api_token'))

    def get_config(self) -> dict:
        # precedence: local cloud config > global profiler cloud config
        config = {}
        try:
            config = Configuration.instance().cloud_config
        except PipeRiderConfigError:
            # piperider may not be initialized
            pass

        if not config.get('default_project'):
            config.update(self.user_profile.get('cloud_config', {}))

        return config if config else {}

    def update_config(self, options: dict):
        cloud_config = self.get_config()
        cloud_config.update(options)
        try:
            Configuration.update_config('cloud_config', cloud_config)
            Configuration.instance().cloud_config = cloud_config
        except FileNotFoundError:
            self.user_profile = update_user_profile({'cloud_config': cloud_config})

    def update_api_token(self):
        if self.api_token:
            update_user_profile({'api_token': self.api_token})
        else:
            update_user_profile({'api_token': None})

    def url(self, uri_path: str):
        if uri_path and not uri_path.startswith('/'):
            uri_path = f'/{uri_path}'

        if self.api_service and self.api_service.endswith('/'):
            return f'{self.api_service[:-1]}{uri_path}'
        return f'{self.api_service}{uri_path}'

    def auth_headers(self):
        return {
            'User-Agent': f'PipeRider CLI/{__version__}',
            'Authorization': f'Bearer {self.api_token}',
        }

    def http_get(self, uri_path):
        try:
            return requests.get(
                self.url(uri_path),
                headers=self.auth_headers()) \
                .json()
        except BaseException:
            return None

    def validate(self) -> (bool, dict):
        if not (self.api_token and self.api_service):
            return False, None

        json_resp = self.http_get('/api/v2/users/me')
        if json_resp and 'email' in json_resp:
            return True, json_resp
        return False, None


class PipeRiderCloud:

    def __init__(self):
        self.service = CloudServiceHelper()
        self.config: dict = self.service.get_config()
        try:
            self.available, self.me = self.service.validate()
        except BaseException:
            self.available = False
            self.me = None

    def update_config(self, options: dict):
        self.service.update_config(options)
        self.config = self.service.get_config()

    def validate(self, api_token=None):
        if api_token:
            self.service.api_token = api_token

        try:
            self.available, self.me = self.service.validate()
        except BaseException:
            self.available = False
            self.me = None

        return self.available

    def logout(self):
        self.service.api_token = None
        self.service.update_api_token()

        # remove the default project when logout
        self.update_config({'default_project': None})

    def magic_signup(self, email: str, username: str = None):
        if self.available:
            return True
        signup_url = self.service.url('/api/v2/credentials/signup')
        payload = {'email': email, 'source': 'cli'}
        if username:
            payload['username'] = username
        response = requests.post(
            signup_url,
            headers={'Content-type': 'application/json', 'Accept': 'text/plain'},
            data=json.dumps(payload)
        )
        if response.status_code == 200:
            return response.json()
        return None

    def magic_login(self, email):
        if self.available:
            return True
        login_url = self.service.url('/api/v2/session/new')
        response = requests.post(
            login_url,
            headers={'Content-type': 'application/json', 'Accept': 'text/plain'},
            data=json.dumps({'email': email, 'source': 'cli'})
        )
        if response.status_code == 200:
            return response.json()
        return None

    def set_default_project(self, project_name):
        self.update_config({'default_project': project_name})

    def get_default_project(self) -> PipeRiderProject:
        if not self.available:
            self.raise_error()

        if self.config.get('default_project'):
            name = self.config.get('default_project')
            project = self.get_project_by_name(name)
            if project:
                return project

        url = self.service.url('/api/v2/workspaces')
        headers = self.service.auth_headers()
        response = requests.get(url, headers=headers)

        if response.status_code != 200:
            return None

        responseData = response.json()

        def genProjects():
            for workspace in responseData.get('data', []):
                for project in workspace.get('projects', []):
                    yield project

        projects = list(genProjects())
        if len(projects) == 1:
            return PipeRiderProject(projects[0])
        else:
            raise PipeRiderNoDefaultProjectError()

    def get_default_workspace_and_project(self):
        if not self.available:
            self.raise_error()

        if self.config.get('default_project'):
            name = self.config.get('default_project')
            workspace_name, project_name = name.split('/')
            return workspace_name, project_name

        # Query workspace
        url = self.service.url('/api/v2/workspaces')
        headers = self.service.auth_headers()
        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            return None, None

        response = response.json()
        if len(response.get('data', [])) == 0:
            return None, None
        workspace = response.get('data')[0]
        workspace_name = workspace.get('name')
        project_name = None
        for p in workspace.get('projects', [])[:1]:
            project_name = p.get('name')
            break

        return workspace_name, project_name

    def list_reports(self, project: PipeRiderProject, datasource=None):
        if not self.available:
            self.raise_error()

        url = self.service.url(f'/api/v2/workspaces/{project.workspace_name}/projects/{project.name}/runs')

        if datasource:
            url = self.service.url(
                f'/api/v2/workspaces/{project.workspace_name}/projects/{project.name}/runs?datasource={datasource}')

        headers = self.service.auth_headers()
        response = requests.get(url, headers=headers)

        if response.status_code != 200:
            return []

        return response.json()

    def get_datasource_id(self, project: PipeRiderProject):
        if not self.available:
            self.raise_error()

        url = self.service.url(f'/api/v2/workspaces/{project.workspace_name}/projects/{project.name}/runs/latest')
        headers = self.service.auth_headers()
        response = requests.get(url, headers=headers)

        if response.status_code != 200:
            return None

        content = response.json()
        return content.get('datasource_id')

    def upload_run(self, file_path, show_progress=True, project: PipeRiderProject = None):
        # TODO validate project name
        if not self.available:
            self.raise_error()

        upload_progress = None
        task_id = None

        def _upload_callback(monitor):
            if show_progress and upload_progress:
                upload_progress.update(task_id, completed=monitor.bytes_read)

        with open(file_path, 'rb') as file:
            encoder = MultipartEncoder(
                fields={'file': ('run.json', file)},
            )
            m = MultipartEncoderMonitor(encoder, _upload_callback)

            url = self.service.url(
                f'/api/v2/workspaces/{project.workspace_name}/projects/{project.name}/runs/upload')
            headers = self.service.auth_headers()
            headers['Content-Type'] = m.content_type

            if show_progress:
                upload_progress = Progress(
                    TextColumn("[bold magenta]{task.description}[/bold magenta]"),
                    BarColumn(bar_width=80),
                    DownloadColumn(binary_units=True),
                    TimeElapsedColumn(),
                )
                task_id = upload_progress.add_task(description=file_path, total=encoder.len)
                upload_progress.start()

            response = requests.post(url, data=m, headers=headers)

            if show_progress:
                upload_progress.stop()

            return response.json()

    def share_run_report(self, workspace_name: str, project_name: str, run_id: int):
        if not self.available:
            self.raise_error()

        url = self.service.url(
            f'/api/v2/workspaces/{workspace_name}/projects/{project_name}/runs/{run_id}/share')
        headers = self.service.auth_headers()
        response = requests.post(url, headers=headers)

        if response.status_code != 200:
            self.raise_error(response.reason)

        return response.json()

    def compare_reports(self, base_id: int, target_id: int, tables_from, project: PipeRiderProject):
        if not self.available:
            self.raise_error()

        url = self.service.url(
            f'/api/v2/workspaces/{project.workspace_name}/projects/{project.name}/runs/{base_id}/compare/{target_id}')
        headers = self.service.auth_headers()
        response = requests.post(url, data=json.dumps({'tables_from': tables_from}), headers=headers)

        if response.status_code != 200:
            self.raise_error(response.reason)

        return response.json()

    def share_compare_report(self, workspace_name: str, project_name: str, base_id: int, target_id: int):
        if not self.available:
            self.raise_error()

        url = self.service.url(
            f'/api/v2/workspaces/{workspace_name}/projects/{project_name}/runs/{base_id}/compare/{target_id}/share')
        headers = self.service.auth_headers()
        response = requests.post(url, headers=headers)

        if response.status_code != 200:
            self.raise_error(response.reason)

        return response.json()

    def raise_error(self, message="Service not available or configuration invalid"):
        raise CloudReportError(message)

    def has_configured(self):
        return self.service.api_token is not None

    def list_projects(self) -> List[dict]:
        if not self.available:
            self.raise_error()

        url = self.service.url('/api/v2/workspaces')
        headers = self.service.auth_headers()
        response = requests.get(url, headers=headers)

        if response.status_code != 200:
            return None

        data = response.json()
        output = []

        if not data.get('success'):
            return None

        data = data.get('data')

        def _parse_projects_with_workspace_list(x):
            for project in x.get('projects', []):
                p = {
                    'id': project.get('id'),
                    'name': project.get('name'),
                    'workspace_name': x.get('name'),
                    'workspace_display_name': x.get('display_name'),
                }
                output.append(p)

        for x in data:
            _parse_projects_with_workspace_list(x)
        return output

    def get_project_by_name(self, name) -> PipeRiderProject:
        workspace_name = None
        if '/' in name:
            workspace_name, project_name = name.split('/')
        else:
            project_name = name

        projects = self.list_projects()
        for project in projects:
            if project.get('name') == project_name and project.get('workspace_name') == workspace_name:
                return PipeRiderProject(project)
        return None


if __name__ == '__main__':
    cloud = PipeRiderCloud()
    cloud.upload_run(json.dumps(dict(foo='barbar')))



import io
import json
import os
import shutil
import sys
from datetime import date, datetime
from typing import Dict, List, Optional

import inquirer
import readchar
from rich.console import Console

import piperider_cli.hack.inquirer as inquirer_hack
from piperider_cli import clone_directory, datetime_to_str, open_report_in_browser, \
    raise_exception_when_directory_not_writable, str_to_datetime
from piperider_cli.configuration import Configuration, ReportDirectory
from piperider_cli.generate_report import setup_report_variables
from piperider_cli.dbt.changeset import SummaryChangeSet
from piperider_cli.dbt.utils import ChangeType


class RunOutput(object):
    def __init__(self, path):
        self.path = path
        self.name = None
        self.created_at = None

        self.table_count = 0
        self.pass_count = 0
        self.fail_count = 0
        self.cloud = None

        try:
            with open(path, 'r') as f:
                run_result = json.load(f)
                self.name = run_result['datasource']['name']
                self.created_at = run_result['created_at']

                tables = run_result.get('tables', {})
                self.table_count = len(tables.keys())

                for test in run_result.get('tests', []):
                    if test.get('status') == 'passed':
                        self.pass_count += 1
                    else:
                        self.fail_count += 1
                self.cloud = run_result.get('cloud')
        except Exception as e:
            if isinstance(e, json.decoder.JSONDecodeError):
                raise json.decoder.JSONDecodeError(
                    f'Invalid JSON in file "{path}"', e.doc, e.pos)
            raise e

    def verify(self) -> bool:
        # TODO: add some verification logic
        return True

    def load(self):
        with open(self.path, 'r') as f:
            data = json.load(f)
        return data

    def refresh(self):
        self.__init__(self.path)

    def __str__(self):
        created_at_str = datetime_to_str(str_to_datetime(self.created_at),
                                         to_tzlocal=True)

        return f'{self.name:12} ' \
               f'#table={self.table_count:<6} ' \
               f'#pass={self.pass_count:<5} ' \
               f'#fail={self.fail_count:<5} ' \
               f'{created_at_str}'


def _merge_keys(base: List[str], target: List[str]):
    '''
    Merge keys from base, target tables. Unlike default union, it preserves the order for column rename, added, removed.

    :param base: keys for base table
    :param target: keys for base table
    :return: merged keys
    '''

    result = []
    while base and target:
        if base[0] == target[0]:
            result.append(base[0])
            base.pop(0)
            target.pop(0)
        elif base[0] in target:
            idx = target.index(base[0])
            for i in target[0:idx]:
                if i not in result:
                    result.append(i)
            result.append(base[0])
            base.pop(0)
            target = target[idx + 1:]
        else:
            result.append(base[0])
            base.pop(0)

    for c in base:
        if c not in result:
            result.append(c)

    for c in target:
        if c not in result:
            result.append(c)

    return result


def join(base, target):
    '''
    Join base and target to a dict which

    keys = (base keys) +  (target keys)
    result[keys] = {base: {...}, target: {...}

    :param base:
    :param target:
    :return:
    '''
    if not base:
        base = dict()
    if not target:
        target = dict()

    keys = _merge_keys(list(base.keys()), list(target.keys()))
    result = dict()
    for key in keys:
        value = dict()
        value['base'] = base.get(key)
        value['target'] = target.get(key)
        result[key] = value
    return result


class ComparisonData(object):
    STATE_ADD = 0
    STATE_DEL = 1
    STATE_MOD = 2

    def __init__(self, base: Dict, target: Dict, tables_from):
        self._id = datetime.now().strftime("%Y%m%d%H%M%S")

        if tables_from == 'target-only':
            target_run_tables = target.get('tables', {}).keys()
            target_run_metrics = {metric.get('name') for metric in target.get('metrics', [])}
            target_run_tests = {test.get('id') for test in target.get('tests', [])}

            base['tables'] = {table_name: base['tables'][table_name] for table_name in base['tables'] if
                              table_name in target_run_tables}
            if 'metrics' in base:
                base['metrics'] = [metric for metric in base['metrics'] if metric.get('name') in target_run_metrics]

            if 'tests' in base:
                base['tests'] = [test for test in base['tests'] if test.get('id') in target_run_tests]
        elif tables_from == 'base-only':
            base_run_tables = base.get('tables', {}).keys()
            base_run_metrics = {metric.get('name') for metric in base.get('metrics', [])}
            base_run_tests = {test.get('id') for test in base.get('tests', [])}

            target['tables'] = {table_name: target['tables'][table_name] for table_name in target['tables'] if
                                table_name in base_run_tables}
            if 'metrics' in target:
                target['metrics'] = [metric for metric in target['metrics'] if metric.get('name') in base_run_metrics]

            if 'tests' in target:
                target['tests'] = [test for test in target['tests'] if test.get('id') in base_run_tests]

        self._base = base
        self._target = target

        self.implicit = []
        self.explicit = []

        self.summary_change_set: Optional[SummaryChangeSet] = None
        self._update_implicit_and_explicit_changeset()

    def _update_implicit_and_explicit_changeset(self):
        try:
            from piperider_cli.dbt.changeset import GraphDataChangeSet
            c = GraphDataChangeSet(self._base, self._target)
            self.explicit = c.list_explicit_changes()
            self.implicit = c.list_implicit_changes()

            self.summary_change_set = SummaryChangeSet(self._base, self._target)
        except BaseException as e:
            console = Console()
            console.print(
                f'[bold yellow]Warning:[/bold yellow] {e}. Got problem to generate changeset.')

    def id(self):
        return self._id

    def to_json(self):
        output = dict(
            created_at=datetime_to_str(datetime.utcnow()),
            base=self._base,
            # TODO: rename input -> target in schema and result json
            input=self._target,
            implicit=self.implicit,
            explicit=self.explicit
        )
        return json.dumps(output, separators=(',', ':'))

    def to_summary_markdown(self):
        out = io.StringIO()

        # Comparison summary
        out.write(self._render_compare_markdown())

        # Per-table Comparison
        base = self._base.get('tables')
        target = self._target.get('tables')

        states = {}
        per_table_out = io.StringIO()
        joined = join(base, target)
        for table_name in joined.keys():
            joined_table = joined[table_name]

            columns_b = joined_table.get('base').get('columns') if joined_table.get('base') else None
            columns_t = joined_table.get('target').get('columns') if joined_table.get('target') else None

            state, result = self._render_table_summary_markdown(table_name, columns_b, columns_t)
            if state == self.STATE_ADD:
                if states.get('added') is None:
                    states['added'] = 0
                states['added'] += 1
            elif state == self.STATE_DEL:
                if states.get('deleted') is None:
                    states['deleted'] = 0
                states['deleted'] += 1
            elif state == self.STATE_MOD:
                if states.get('schema changed') is None:
                    states['schema changed'] = 0
                states['schema changed'] += 1
            per_table_out.write(result)

        # Per-table summary
        states = sorted(states.items())
        table_summary_hint = ', '.join([f'{state}={num}' for state, num in states])
        if table_summary_hint:
            table_summary_hint = f' ({table_summary_hint})'

        out.write("<details>\n")
        out.write(f"<summary>Tables Summary{table_summary_hint}</summary>\n")
        out.write("<blockquote>\n")
        out.write("\n")
        out.write(per_table_out.getvalue())
        out.write("</blockquote></details>")

        base = self._base.get('metrics', [])
        target = self._target.get('metrics', [])
        out.write(self._render_metrics_comparison_markdown(base, target))

        return out.getvalue()

    def to_summary_markdown_ng(self):
        if self._base.get('dbt') is None or self._target.get('dbt') is None:
            console = Console()
            console.print("[bold yellow]Warning: [/bold yellow]'summary.md' report is not generated.")
            console.print("To generate a summary.md file, please run the 'piperider run' command in a dbt project "
                          "and use the latest version of piperider.")
            return ""

        # TODO replace to new generator
        if self.summary_change_set:
            return self.summary_change_set.generate_markdown()

        return ""

    @staticmethod
    def _value_with_annotation(key, annotation=None):
        annotation_str = f" ({annotation})" if annotation else ''
        return f"{key}{annotation_str}"

    @staticmethod
    def _value_with_change(base, target):
        if target is None:
            return f"~~{base}~~"
        elif base is None:
            return target
        elif base != target:
            return f"~~{base}~~<br/>{target}"
        else:
            return target

    @staticmethod
    def _value_with_delta(base, target, percentage=False):
        if target is None:
            return '-'

        annotation = '%' if percentage else ''
        delta = ''
        if base is not None:
            diff = target - base
            if percentage:
                diff *= 100
            delta = f" ({'+' if target >= base else ''}{round(diff, 2)}{annotation})"

        if percentage:
            target = round(target * 100, 2)

        return f"{target}{annotation}{delta}"

    @staticmethod
    def _display_metrics_delta(b_value, t_value):
        if b_value == '' or t_value == '':
            return ''
        elif isinstance(b_value, str) and b_value.startswith('~~'):
            return ''
        elif isinstance(t_value, str) and t_value.startswith('~~'):
            return ''
        elif b_value == t_value:
            return '-'
        b_value = 0 if b_value == '-' else b_value
        t_value = 0 if t_value == '-' else t_value
        delta = t_value - b_value

        return f"{'+' if delta >= 0 else ''}{delta}"

    @staticmethod
    def _get_column_changed(base, target):
        columns_b = base.get('columns') if base else None
        columns_t = target.get('columns') if target else None
        joined = join(columns_b, columns_t)
        added = 0
        deleted = 0
        changed = 0

        for column_name in joined.keys():
            joined_column = joined[column_name]
            b = joined_column.get('base')
            t = joined_column.get('target')

            schema_type_b = b.get('schema_type') if b else None
            schema_type_t = t.get('schema_type') if t else None

            if b is None:
                added += 1
            elif t is None:
                deleted += 1
            elif schema_type_b != schema_type_t:
                changed += 1

        return {
            'added': added,
            'deleted': deleted,
            'changed': changed,
        }

    def _render_compare_markdown(self):
        base = self._base.get('tables')
        target = self._target.get('tables')

        out = io.StringIO()
        joined = join(base, target)
        print("Table | Rows | Columns ", file=out)
        print("--- | --- | ---", file=out)
        for table_name in joined.keys():
            joined_table = joined[table_name]
            b = joined_table.get('base')
            t = joined_table.get('target')

            annotation = None
            change_message_str = ''

            if b is None:
                annotation = '+'
            elif t is None:
                annotation = '-'
            else:
                column_changed = self._get_column_changed(b, t)
                messages = []

                if column_changed['added'] > 0:
                    annotation = '!'
                    messages.append(f"added={column_changed['added']}")
                if column_changed['deleted'] > 0:
                    annotation = '!'
                    messages.append(f"deleted={column_changed['deleted']}")
                if column_changed['changed'] > 0:
                    annotation = '!'
                    messages.append(f"changed={column_changed['changed']}")

                if messages:
                    change_message_str = ', '.join(messages)
                    change_message_str = f"<br/>({change_message_str})"

            rows_b, cols_b = (b.get('row_count'), b.get('col_count')) if b else (None, None)
            rows_t, cols_t = (t.get('row_count'), t.get('col_count')) if t else (None, None)

            out.write(f"{self._value_with_annotation(table_name, annotation)} | ")
            out.write(f"{self._value_with_delta(rows_b, rows_t)} | ")
            out.write(f"{self._value_with_delta(cols_b, cols_t)}{change_message_str} \n")

        base = self._base.get('metrics', [])
        target = self._target.get('metrics', [])
        base_metrics = {base_metric.get('name'): base_metric for base_metric in base}
        target_metrics = {target_metric.get('name'): target_metric for target_metric in target}

        joined = join(base_metrics, target_metrics)
        out_metrics = io.StringIO()
        if joined:
            def _display_period_annotation(period_grain):
                if period_grain == 'day':
                    return '(Yesterday)'
                else:
                    return f'(Last {period_grain})'

            print("Metric | Period | Base | Target | +/- ", file=out_metrics)
            print("--- | --- | :-: | :-: | :-: ", file=out_metrics)
            for metric_name in joined.keys():
                joined_metric = joined[metric_name]
                b = joined_metric.get('base')
                t = joined_metric.get('target')
                b_data = {}
                t_data = {}
                if b:
                    grain = b.get('grain')
                    metric_label = b.get('label')
                    b_data = {row[0]: row[1] for row in b.get('data')}
                if t:
                    grain = t.get('grain')
                    metric_label = t.get('label')
                    t_data = {row[0]: row[1] for row in t.get('data')}

                t_dates = t_data.keys() if t_data.keys() else b_data.keys()
                t_dates = sorted(list(t_dates), key=lambda k: date.fromisoformat(k), reverse=True)
                t_last_date = t_dates[1]

                if t_last_date not in b_data:
                    b_result = ''
                else:
                    b_result = b_data.get(t_last_date) if b_data.get(t_last_date) is not None else '-'

                if t_last_date not in t_data:
                    t_result = ''
                else:
                    t_result = t_data.get(t_last_date) if t_data.get(t_last_date) is not None else '-'

                out_metrics.write(f"{metric_label} | {t_last_date} {_display_period_annotation(grain)} | ")
                out_metrics.write(f"{b_result} | {t_result} | {self._display_metrics_delta(b_result, t_result)} \n ")

        return f"""<details>
<summary>Comparison Summary</summary>

{out.getvalue()}
{out_metrics.getvalue()}
</details>
"""

    def _render_table_summary_markdown(self, table_name, base, target):
        out = io.StringIO()
        joined = join(base, target)
        print("Column | Type | Valid % | Distinct %", file=out)
        print("--- | --- | --- | ---", file=out)
        table_modified = False
        for column_name in joined.keys():
            joined_column = joined[column_name]
            b = joined_column.get('base')
            t = joined_column.get('target')

            schema_type_b = self._get_metric_from_report(b, 'schema_type', None)
            valids_p_b = self._get_metric_from_report(b, 'valids_p', None)
            distinct_p_b = self._get_metric_from_report(b, 'distinct_p', None)

            schema_type_t = self._get_metric_from_report(t, 'schema_type', None)
            valids_p_t = self._get_metric_from_report(t, 'valids_p', None)
            distinct_p_t = self._get_metric_from_report(t, 'distinct_p', None)

            annotation = None
            if b is None:
                annotation = '+'
                table_modified = True
            elif t is None:
                annotation = '-'
                table_modified = True
            elif schema_type_b != schema_type_t:
                annotation = '!'
                table_modified = True

            out.write(f"{self._value_with_annotation(column_name, annotation)} | ")
            out.write(f"{self._value_with_change(schema_type_b, schema_type_t)} | ")
            out.write(f"{self._value_with_delta(valids_p_b, valids_p_t, percentage=True)} | ")
            out.write(f"{self._value_with_delta(distinct_p_b, distinct_p_t, percentage=True)} \n")

        annotation = ''
        state = None
        if base is None:
            annotation = ' (+)'
            state = self.STATE_ADD
        elif target is None:
            annotation = ' (-)'
            state = self.STATE_DEL
        elif table_modified:
            annotation = ' (!)'
            state = self.STATE_MOD

        return state, f"""<details>
<summary>{table_name}{annotation}</summary>

{out.getvalue()}
</details>
"""

    @staticmethod
    def _get_metric_from_report(report, metric, default):
        return report.get(metric, default) if report else default

    def _render_metrics_comparison_markdown(self, base, target):
        out = io.StringIO()

        base_metrics = {base_metric.get('name'): base_metric for base_metric in base}
        target_metrics = {target_metric.get('name'): target_metric for target_metric in target}
        joined = join(base_metrics, target_metrics)
        if joined:
            out_metrics = io.StringIO()
            states = {}
            for metric_name in joined.keys():
                notation = ''
                metric_warn = False
                joined_metric = joined[metric_name]
                b = joined_metric.get('base')
                t = joined_metric.get('target')
                date_grain = 'date'
                b_data = {}
                t_data = {}
                if b:
                    date_grain = b.get('headers')[0]
                    date_grain = date_grain[0].upper() + date_grain[1:]
                    metric_label = b.get('label')
                    b_data = {row[0]: row[1] for row in b.get('data')}
                else:
                    notation = '(+)'
                    if states.get('added') is None:
                        states['added'] = 0
                    states['added'] += 1
                if t:
                    date_grain = t.get('headers')[0]
                    date_grain = date_grain[0].upper() + date_grain[1:]
                    metric_label = t.get('label')
                    t_data = {row[0]: row[1] for row in t.get('data')}
                else:
                    notation = '(-)'
                    if states.get('deleted') is None:
                        states['deleted'] = 0
                    states['deleted'] += 1

                b_dates = b_data.keys() if b_data.keys() else t_data.keys()
                b_dates = sorted(list(b_dates), key=lambda k: date.fromisoformat(k), reverse=True)
                b_latest_date = b_dates[0]
                t_dates = t_data.keys() if t_data.keys() else b_data.keys()
                t_dates = sorted(list(t_dates), key=lambda k: date.fromisoformat(k), reverse=True)
                t_latest_date = t_dates[0]

                out_metric = io.StringIO()
                for d in t_dates:
                    if d not in b_data:
                        b_result = ''
                    else:
                        b_result = b_data.get(d) if b_data.get(d) is not None else '-'
                        if date.fromisoformat(d) == date.fromisoformat(b_latest_date):
                            b_result = f'~~{b_data.get(d)}~~' if b_data.get(d) is not None else '-'

                    if d not in t_data:
                        t_result = ''
                    else:
                        t_result = t_data.get(d) if t_data.get(d) is not None else '-'
                        if date.fromisoformat(d) == date.fromisoformat(t_latest_date):
                            t_result = f'~~{t_data.get(d)}~~' if t_data.get(d) is not None else '-'

                    if date.fromisoformat(t_latest_date) >= date.fromisoformat(d) >= date.fromisoformat(b_latest_date):
                        delta = ''
                    else:
                        delta = self._display_metrics_delta(b_result, t_result)
                        if delta != '-' and delta != '':
                            metric_warn = True
                    out_metric.write(f"{d} | {b_result} | {t_result} | {delta}\n")

                if metric_warn:
                    notation = '(!)'
                    if states.get('metric changed') is None:
                        states['metric changed'] = 0
                    states['metric changed'] += 1

                out_metrics.write("<details>\n")
                out_metrics.write(f"<summary>{metric_label} {notation}</summary>\n\n")
                if not t:
                    out_metrics.write("Metric is not available in target\n")
                else:
                    out_metrics.write(f"{date_grain} | Base | Target | -/+ \n")
                    out_metrics.write(":-: | :-: | :-: | :-: \n")
                    out_metrics.write(out_metric.getvalue())
                out_metrics.write("</details>\n")

            states = sorted(states.items())
            metrics_summary_hint = ', '.join([f'{state}={num}' for state, num in states])
            if metrics_summary_hint:
                metrics_summary_hint = f' ({metrics_summary_hint})'
            out.write("<details>\n")
            out.write(f"<summary>Metrics Summary{metrics_summary_hint}</summary>\n")
            out.write("<blockquote>\n\n")
            out.write(out_metrics.getvalue())
            out.write("</blockquote></details>")

        return out.getvalue()

    def to_cli_stats(self, console):
        console.print()

        if self.summary_change_set is None:
            return

        console.print("Statistics:")

        for d in [self.summary_change_set.models, self.summary_change_set.metrics]:
            output = [f"  {d.resource_type}: total={d.total}, explict={d.explicit_changes}",
                      f"(added={len([x for x in d.explicit_changeset if x.change_type == ChangeType.ADDED])}, "
                      f"removed={len([x for x in d.explicit_changeset if x.change_type == ChangeType.REMOVED])}, "
                      f"modified={len([x for x in d.explicit_changeset if x.change_type == ChangeType.MODIFIED])}), ",
                      f"impacted={d.impacted}, implicit={d.implicit_changes}"]

            console.print("".join(output))

        console.print("")


def prepare_default_output_path(filesystem: ReportDirectory, created_at):
    latest_symlink_path = os.path.join(filesystem.get_comparison_dir(), 'latest')
    latest_source = created_at
    comparison_path = os.path.join(filesystem.get_comparison_dir(), created_at)

    if not os.path.exists(comparison_path):
        os.makedirs(comparison_path, exist_ok=True)

    # Create a symlink pointing to the latest comparison directory
    if os.path.islink(latest_symlink_path):
        os.unlink(latest_symlink_path)

    console = Console()
    if not os.path.exists(latest_symlink_path):
        try:
            os.symlink(latest_source, latest_symlink_path)
        except OSError as e:
            """
            System Error Codes:
                ERROR_PRIVILEGE_NOT_HELD
                1314 (0x522)
                A required privilege is not held by the client.
                https://learn.microsoft.com/en-us/windows/win32/debug/system-error-codes--1300-1699-
            """
            if e.winerror is not None and e.winerror == 1314:
                console.print(
                    f'[bold yellow]Warning:[/bold yellow] {e}. To solve this, run piperider as an administrator')
            else:
                raise e
    else:
        console.print(f'[bold yellow]Warning: {latest_symlink_path} already exists[/bold yellow]')

    return comparison_path


class CompareReport(object):
    def __init__(self, profiler_output_path, a=None, b=None, datasource=None, profiler_outputs=None):
        self.profiler_output_path = profiler_output_path
        self.profiler_outputs = profiler_outputs
        self.console = Console()
        self.a: RunOutput = RunOutput(a) if a else None
        self.b: RunOutput = RunOutput(b) if b else None
        self.datasource = datasource

    def list_existing_outputs(self, output_search_path=None) -> List[RunOutput]:
        """
        List existing profiler outputs.
        """

        def _walk_throw_runs(path):
            outputs = []
            for root, dirs, _ in os.walk(path):
                for dir in dirs:
                    if dir == 'latest':
                        continue
                    run_json = os.path.join(root, dir, 'run.json')
                    if not os.path.exists(run_json):
                        continue
                    output = RunOutput(run_json)
                    if self.datasource and output.name != self.datasource:
                        continue
                    outputs.append(output)
            outputs.sort(key=lambda x: (x.name, x.created_at), reverse=True)
            return outputs

        if output_search_path is None:
            output_search_path = self.profiler_output_path

        return _walk_throw_runs(output_search_path)

    def get_the_last_two_reports(self):
        outputs = self.list_existing_outputs()
        outputs.sort(key=lambda x: x.created_at)
        outputs = list(filter(lambda x: x.name == outputs[-1].name, outputs))
        if len(outputs) < 2:
            return None, None
        return outputs[-2:]

    def select_reports(self, use_last_two=None, reverse=False):
        if use_last_two:
            first, second = self.get_the_last_two_reports()
            if reverse:
                self.a, self.b = second, first
            else:
                self.a, self.b = first, second
        elif self.a is None and self.b is None:
            self.a, self.b = self.select_two_reports()
        elif self.a and self.b is None:
            self.b = self.select_one_report()
            pass
        elif self.a is None and self.b:
            self.a = self.select_one_report()

        if self.a and self.b:
            self.console.print('Selected reports:')
            self.console.print(f'  Base:   {self.a.path}', soft_wrap=True)
            self.console.print(f'  Target: {self.b.path}', soft_wrap=True)
            return True
        return False

    def select_multiple_reports(self, action='compare', limit=None):
        def _report_validater(answers, current) -> bool:
            if limit is None:
                return len(current) > 0
            else:
                return len(current) == limit

        profiler_outputs = self.profiler_outputs if self.profiler_outputs is not None else self.list_existing_outputs()
        arrow_alias_msg = ''
        if sys.platform == "win32" or sys.platform == "cygwin":
            # change readchar key UP & DOWN by 'w' and 's'
            readchar.key.UP = 'w'
            readchar.key.DOWN = 's'
            arrow_alias_msg = " 'w' to Up, 's' to Down,"

        if limit is None:
            if len(profiler_outputs) == 0:
                raise Exception("Not enough reports to compare. Please run 'piperider run' first.")
            questions = [
                inquirer.Checkbox('profiler_output',
                                  message=f"Please select the reports to {action} ({arrow_alias_msg} SPACE to select, and ENTER to confirm )",
                                  choices=profiler_outputs,
                                  carousel=True,
                                  validate=_report_validater,
                                  )
            ]
        else:
            if len(profiler_outputs) < limit:
                raise Exception("Not enough reports to compare. Please run 'piperider run' first.")
            report_msg = 'a report' if limit == 1 else f'the {limit} reports'
            questions = [
                inquirer_hack.LimitedCheckboxQuestion('profiler_output',
                                                      message=f"Please select {report_msg} to {action} ({arrow_alias_msg} SPACE to select, and ENTER to confirm )",
                                                      choices=profiler_outputs,
                                                      carousel=True,
                                                      validate=_report_validater,
                                                      limited=limit,
                                                      )
            ]

        answers = inquirer_hack.prompt_ex(questions, raise_keyboard_interrupt=True)
        if answers:
            return answers['profiler_output']
        return None

    def select_one_report(self, action='compare') -> RunOutput:
        answers = self.select_multiple_reports(action=action, limit=1)
        if answers:
            return answers[0]
        else:
            return None

    def select_two_reports(self, action='compare') -> (RunOutput, RunOutput):
        """
        Select multiple files from a list of files.
        """

        answers = self.select_multiple_reports(action=action, limit=2)
        if answers:
            return answers[0], answers[1]
        return None, None

    def generate_data(self, tables_from: str) -> ComparisonData:
        if self.a is None or self.b is None:
            raise Exception("Please select reports to compare first.")

        return ComparisonData(self.a.load(), self.b.load(), tables_from)

    @staticmethod
    def exec(*, a=None, b=None, last=None, datasource=None, report_dir=None, output=None, tables_from='all',
             summary_file=None, force_upload=False, enable_share=False, open_report=False, project_name: str = None,
             debug=False, show_progress=False, reverse_last=False):
        console = Console()
        console.rule('Comparison report', style='bold blue')

        filesystem = Configuration.instance().activate_report_directory(report_dir=report_dir)
        raise_exception_when_directory_not_writable(output)

        report = CompareReport(filesystem.get_output_dir(), a, b, datasource=datasource)
        if not report.select_reports(use_last_two=last, reverse=reverse_last):
            raise Exception('No valid reports found')

        from piperider_cli.cloud_connector import CloudConnector
        report_url = None
        summary_data = None

        if force_upload and (report.a.cloud is None or report.a.cloud.get('project_name') != project_name):
            CloudConnector.upload_report(report.a.path, show_progress=show_progress, project_name=project_name)
            report.a.refresh()

        if force_upload and (report.b.cloud is None or report.b.cloud.get('project_name') != project_name):
            CloudConnector.upload_report(report.b.path, show_progress=show_progress, project_name=project_name)
            report.b.refresh()

        # Generate comparison report URL & summary markdown
        if report.a.cloud and report.b.cloud:
            base = str(report.a.cloud.get('run_id'))
            target = str(report.b.cloud.get('run_id'))
            project_name = report.a.cloud.get('project_name')
            response = CloudConnector.generate_compare_report(base, target, project_name=project_name, debug=False)
            if response:
                report_url = response.get('url')
                summary_data = response.get('summary')

        comparison_data = report.generate_data(tables_from)

        from piperider_cli import data
        report_template_dir = os.path.join(os.path.dirname(data.__file__), 'report', 'comparison-report')
        with open(os.path.join(report_template_dir, 'index.html')) as f:
            report_template_html = f.read()

        def output_report(directory):
            clone_directory(report_template_dir, directory)
            filename = os.path.join(directory, 'index.html')
            with open(filename, 'w') as f:
                html = setup_report_variables(report_template_html, False, comparison_data.to_json())
                f.write(html)

        def output_summary(directory, summary_data):
            filename = os.path.join(directory, 'summary.md')
            with open(filename, 'w') as f:
                f.write(summary_data)

        data_id = comparison_data.id()
        summary_data = summary_data if summary_data else comparison_data.to_summary_markdown_ng()

        default_report_directory = prepare_default_output_path(filesystem, data_id)
        output_report(default_report_directory)
        if summary_data:
            output_summary(default_report_directory, summary_data)

        comparison_dir = filesystem.get_comparison_dir()
        report_path = os.path.join(comparison_dir, 'latest', 'index.html')
        summary_md_path = os.path.join(comparison_dir, 'latest', 'summary.md')
        if not os.path.exists(os.path.join(comparison_dir, 'latest')):
            report_path = os.path.join(default_report_directory, 'index.html')
            summary_md_path = os.path.join(default_report_directory, 'summary.md')

        if enable_share:
            if report.a.cloud is None or report.b.cloud is None:
                console.print(
                    '[[bold yellow]Skip[/bold yellow]] Please enable cloud auto upload or use "piperider compare --upload" to upload reports to cloud first.')
            else:
                from piperider_cli.cloud_connector import CloudConnector
                base = str(report.a.cloud.get('run_id'))
                target = str(report.b.cloud.get('run_id'))
                CloudConnector.share_compare_report(base, target)

        if output:
            clone_directory(default_report_directory, output)
            report_path = os.path.abspath(os.path.join(output, 'index.html'))
            summary_md_path = os.path.abspath(os.path.join(output, 'summary.md'))

        if summary_file:
            summary_file = os.path.abspath(summary_file)
            summary_dir = os.path.dirname(summary_file)
            if summary_dir:
                os.makedirs(summary_dir, exist_ok=True)
            shutil.copyfile(summary_md_path, summary_file)
            summary_md_path = summary_file

        comparison_data.to_cli_stats(console)
        console.print()
        console.print(f"Comparison report: {report_path}")
        if summary_data:
            console.print(f"Comparison summary: {summary_md_path}")
        if report_url:
            console.print(f"Comparison report URL: {report_url}", soft_wrap=True)

        if open_report:
            if report_url:
                open_report_in_browser(report_url, True)
            else:
                open_report_in_browser(report_path)

        if debug:
            # Write comparison data to file
            with open(os.path.join(default_report_directory, 'comparison_data.json'), 'w') as f:
                f.write(comparison_data.to_json())



def disable_dbt_compile_stats():
    import dbt.compilation as compilation

    class context_class:

        def __enter__(self):
            self.original_print_compile_stats = compilation.print_compile_stats
            compilation.print_compile_stats = lambda x: None

        def __exit__(self, exc_type, exc_val, exc_tb):
            compilation.print_compile_stats = self.original_print_compile_stats

    try:
        from dbt.events.functions import cleanup_event_logger
        cleanup_event_logger()
    except ImportError:
        # we dont support the dbt version less than 1.4
        pass

    return context_class()



import json
import os
import re
import shutil
from base64 import b64encode
from typing import Dict

from rich.console import Console

from piperider_cli import __version__, open_report_in_browser, sentry_dns, sentry_env, event, get_run_json_path
from piperider_cli import clone_directory, raise_exception_when_directory_not_writable
from piperider_cli.configuration import Configuration
from piperider_cli.error import PipeRiderNoProfilingResultError


def prepare_piperider_metadata():
    configuration = Configuration.instance()
    project_id = configuration.get_telemetry_id()
    metadata = {
        'name': 'PipeRider',
        'sentry_env': sentry_env,
        'sentry_dns': sentry_dns,
        'version': __version__,
        'amplitude_api_key': event._get_api_key(),
        'amplitude_user_id': event._collector._user_id,
        'amplitude_project_id': project_id,
    }
    return metadata


def _validate_input_result(result):
    for f in ['tables', 'id', 'created_at', 'datasource']:
        if f not in result:
            return False
    return True


def setup_report_variables(template_html: str, is_single: bool, data):
    if isinstance(data, dict):
        output = json.dumps(data)
    else:
        output = data
    metadata = json.dumps(prepare_piperider_metadata())
    encoded_output = b64encode(bytes(output, "utf-8")).decode("ascii")
    if is_single:
        variables = f'<script id="piperider-report-variables">\n' \
                    f'window.PIPERIDER_METADATA={metadata};' \
                    f'window.PIPERIDER_SINGLE_REPORT_DATA=JSON.parse(atob("{encoded_output}"));' \
                    f'window.PIPERIDER_COMPARISON_REPORT_DATA="";</script>'
    else:
        variables = f'<script id="piperider-report-variables">\n' \
                    f'window.PIPERIDER_METADATA={metadata};' \
                    f'window.PIPERIDER_SINGLE_REPORT_DATA="";' \
                    f'window.PIPERIDER_COMPARISON_REPORT_DATA=JSON.parse(atob("{encoded_output}"));</script>'
    html_parts = re.sub(r'<script id="piperider-report-variables">.+?</script>', '#PLACEHOLDER#', template_html).split(
        '#PLACEHOLDER#')
    html = html_parts[0] + variables + html_parts[1]
    return html


def _generate_static_html(result, html, output_path):
    filename = os.path.join(output_path, "index.html")
    with open(filename, 'w') as f:
        html = setup_report_variables(html, True, result)
        f.write(html)


class GenerateReport:
    @staticmethod
    def exec(input=None, report_dir=None, output=None, open_report=None, open_in_cloud=None):
        filesystem = Configuration.instance().activate_report_directory(report_dir)
        raise_exception_when_directory_not_writable(output)

        console = Console()

        from piperider_cli import data
        report_template_dir = os.path.join(os.path.dirname(data.__file__), 'report', 'single-report')
        with open(os.path.join(report_template_dir, 'index.html')) as f:
            report_template_html = f.read()

        run_json_path = get_run_json_path(filesystem.get_output_dir(), input)
        if os.path.islink(os.path.dirname(run_json_path)):
            run_json_path = os.path.realpath(run_json_path)
        if not os.path.isfile(run_json_path):
            print(os.path.abspath(run_json_path))
            raise PipeRiderNoProfilingResultError(run_json_path)

        with open(run_json_path) as f:
            result = json.loads(f.read())
        if not _validate_input_result(result):
            console.print(f'[bold red]Error: {run_json_path} is invalid[/bold red]')
            return

        console.print('')
        console.print(f'Generating reports from: {run_json_path}', soft_wrap=True)

        def output_report(target_directory):
            clone_directory(report_template_dir, target_directory)
            _generate_static_html(result, report_template_html, target_directory)

        # output the report to the default directory (same with the run.json)
        default_output_directory = os.path.dirname(run_json_path)
        output_report(default_output_directory)

        if output:
            output_report(output)
            shutil.copyfile(run_json_path, os.path.join(output, os.path.basename(run_json_path)))
            console.print(
                f"Report generated in: {os.path.join(output, 'index.html')}", soft_wrap=True)
        else:
            console.print(
                f"Report generated in: {os.path.join(default_output_directory, 'index.html')}", soft_wrap=True)

        # only open the local file report if auto-upload is OFF
        if open_report and not open_in_cloud:
            result_output = f"{os.path.abspath(output) if output else default_output_directory}"
            result_output = os.path.join(result_output, 'index.html')
            open_report_in_browser(result_output)



from datetime import datetime

from sqlalchemy import select, MetaData, Table, func

from piperider_cli.assertion_engine import AssertionContext, AssertionResult
from piperider_cli.assertion_engine.assertion import ValidationResult
from piperider_cli.assertion_engine.types.base import BaseAssertionType
from piperider_cli.assertion_engine.types.assert_metrics import AssertMetric


class AssertColumnNotNull(BaseAssertionType):
    def name(self):
        return "assert_column_not_null"

    def execute(self, context: AssertionContext):
        return assert_column_not_null(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        return ValidationResult(context).keep_no_args()


class AssertColumnNull(BaseAssertionType):
    def name(self):
        return "assert_column_null"

    def execute(self, context: AssertionContext):
        return assert_column_null(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        return ValidationResult(context).keep_no_args()


class AssertColumnUnique(BaseAssertionType):
    def name(self):
        return "assert_column_unique"

    def execute(self, context: AssertionContext):
        return assert_column_unique(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        return ValidationResult(context).keep_no_args()


class AssertColumnExist(BaseAssertionType):
    def name(self):
        return "assert_column_exist"

    def execute(self, context: AssertionContext):
        return assert_column_exist(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        return ValidationResult(context).keep_no_args()


class AssertColumnValue(BaseAssertionType):
    def __init__(self):
        self.set_op = 'in'
        self.range_ops = ['gte', 'lte', 'gt', 'lt']

    def name(self):
        return "assert_column_value"

    def execute(self, context: AssertionContext):
        table = context.table
        column = context.column
        metrics = context.profiler_result

        target_metrics = None
        if metrics:
            target_metrics = metrics.get('tables', {}).get(table)
            if column:
                target_metrics = target_metrics.get('columns', {}).get(column)

            if not target_metrics:
                return context.result.fail_with_metric_not_found_error(context.table, context.column)

        assert_ops = list(context.asserts.keys())
        if assert_ops[0] in self.range_ops:
            return self._assert_column_value_range(context, target_metrics)
        elif assert_ops[0] == self.set_op:
            return self._assert_column_value_set(context, target_metrics)

        return context.result.fail()

    def validate(self, context: AssertionContext) -> ValidationResult:
        results = ValidationResult(context)

        ops = [self.set_op] + self.range_ops

        if not context.asserts:
            results.errors.append(f'At least one of {ops} is needed.')
            return results

        if not isinstance(context.asserts, dict):
            results.errors.append('assert should be an associative array')
            return results

        results = results.allow_only(*ops)
        if results.errors:
            return results

        assert_ops = list(context.asserts.keys())
        if assert_ops[0] in self.range_ops:
            results = results.require_metric_consistency(*self.range_ops)
            if results.errors:
                return results

            self._assert_value_validation(context.asserts, results)
        elif assert_ops[0] == self.set_op:
            if not isinstance(context.asserts.get(self.set_op), list):
                results.errors.append(f"'{self.set_op}' should specify a list")
                return results
            results = results.require_same_types(self.set_op)

        return results

    @staticmethod
    def _assert_value_validation(value_boundary: dict, results: ValidationResult):
        if len(value_boundary.keys()) == 1:
            pass
        elif len(value_boundary.keys()) == 2:
            lower = None
            upper = None
            for op, v in value_boundary.items():
                if op.startswith('lt'):
                    upper = v
                elif op.startswith('gt'):
                    lower = v

            if upper is None or lower is None:
                results.errors.append('Please specified your metric upper and lower boundary')
                return

            if isinstance(upper, str) and isinstance(lower, str):
                upper = datetime.fromisoformat(upper)
                lower = datetime.fromisoformat(lower)
            if upper < lower:
                results.errors.append("The 'lt' or 'lte' value should be greater than or equal to "
                                      "the 'gt' or 'gte' value.")
        else:
            results.errors.append('The number of operator should be 1 or 2.')

    @staticmethod
    def _assert_column_value_range(context, target_metrics):
        context.result.expected = AssertMetric.to_interval_notation(context.asserts)
        has_fail = False

        if target_metrics.get('min') is not None:
            if not AssertMetric.assert_metric_boundary(target_metrics.get('min'), context.asserts):
                has_fail = True
        else:
            return context.result.fail()

        if target_metrics.get('max') is not None:
            if not AssertMetric.assert_metric_boundary(target_metrics.get('max'), context.asserts):
                has_fail = True
        else:
            return context.result.fail()

        context.result.actual = AssertMetric.to_interval_notation({
            'gte': target_metrics.get('min'),
            'lte': target_metrics.get('max')
        })

        if has_fail:
            return context.result.fail()

        return context.result.success()

    @staticmethod
    def _assert_column_value_set(context, target_metrics):
        assert_set = list(context.asserts.values())[0]
        context.result.expected = assert_set
        assert_set = set(assert_set)

        samples = None
        if target_metrics:
            samples = target_metrics.get('samples')
            distinct = target_metrics.get('distinct')
            topk = []
            if target_metrics.get('topk'):
                topk = target_metrics.get('topk').get('values', [])

            if distinct and distinct > len(assert_set):
                subtraction = AssertColumnValue._column_value_set_subtraction(context, target_metrics, None)
                context.result.actual = AssertColumnValue._get_column_value_set_actual_msg(subtraction)
                return context.result.fail()

            # TODO: define topk default max length
            if distinct and distinct <= 50 and topk:
                if len(assert_set) < len(topk):
                    subtraction = AssertColumnValue._column_value_set_subtraction(context, None, topk)
                    context.result.actual = AssertColumnValue._get_column_value_set_actual_msg(subtraction)
                    return context.result.fail()

                for k in topk:
                    if k not in assert_set:
                        subtraction = AssertColumnValue._column_value_set_subtraction(context, None, topk)
                        context.result.actual = AssertColumnValue._get_column_value_set_actual_msg(subtraction)
                        return context.result.fail()

                return context.result.success()

        # try to query 4 additional categories
        # if there's additional categories, it means the dataset has more data category than assertion
        # also we show up to 3 items in the category subtraction
        result = AssertColumnValue._query_column_category(context, samples, len(assert_set) + 4)
        if len(result) > len(assert_set):
            subtraction = AssertColumnValue._column_value_set_subtraction(context, None, result)
            context.result.actual = AssertColumnValue._get_column_value_set_actual_msg(subtraction)
            return context.result.fail()

        for row in result:
            if row not in assert_set:
                subtraction = AssertColumnValue._column_value_set_subtraction(context, None, result)
                context.result.actual = AssertColumnValue._get_column_value_set_actual_msg(subtraction)
                return context.result.fail()

        return context.result.success()

    @staticmethod
    def _query_column_category(context, samples, size):
        table = context.table
        column = context.column

        metadata = MetaData()
        Table(table, metadata, autoload_with=context.engine)
        t = metadata.tables[table]
        c = t.columns[column]

        with context.engine.connect() as conn:
            base = select(c.label('c')).select_from(t)
            if samples:
                base = base.limit(samples)
            cte = base.cte()
            stmt = select(
                func.distinct(cte.c.c)
            ).select_from(
                cte
            ).where(
                cte.c.c.isnot(None)
            )
            results = conn.execute(stmt).fetchmany(size=size)

        return [r[0] for r in results]

    @staticmethod
    def _column_value_set_subtraction(context, target_metrics, actual_set):
        assert_set = set(list(context.asserts.values())[0])
        samples = None
        if target_metrics:
            samples = target_metrics.get('samples')
            distinct = target_metrics.get('distinct')
            topk = []
            if target_metrics.get('topk'):
                topk = target_metrics.get('topk').get('values', [])

            if distinct and distinct <= 50 and topk:
                actual_set = topk

        if actual_set:
            return set(actual_set) - assert_set

        # try to query 4 additional categories
        # if there's additional categories, it means the dataset has more data category than assertion
        # also we show up to 3 items in the category subtraction
        result = AssertColumnValue._query_column_category(context, samples, len(assert_set) + 4)

        return set(result) - assert_set

    @staticmethod
    def _get_column_value_set_actual_msg(subtraction):
        subtraction = list(subtraction)
        if len(subtraction) == 1:
            actual_msg = f'{subtraction} is in the set'
        elif len(subtraction) <= 3:
            actual_msg = f'{subtraction} are in the set'
        else:
            actual_msg = f'{subtraction[0:3]} and more are in the set'

        return actual_msg


def assert_column_not_null(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    column_metrics = metrics.get('tables', {}).get(table, {}).get('columns', {}).get(column)
    if not column_metrics:
        return context.result.fail_with_metric_not_found_error(context.table, context.column)

    if context.asserts:
        return context.result.fail_with_no_assert_is_required()

    samples = column_metrics.get('samples')
    non_nulls = column_metrics.get('non_nulls')

    success = (samples == non_nulls)
    context.result.actual = None

    if success:
        return context.result.success()

    return context.result.fail()


def assert_column_null(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    column_metrics = metrics.get('tables', {}).get(table, {}).get('columns', {}).get(column)
    if not column_metrics:
        # cannot find the column in the metrics
        return context.result.fail_with_metric_not_found_error(context.table, context.column)

    if context.asserts:
        return context.result.fail_with_no_assert_is_required()

    non_nulls = column_metrics.get('non_nulls')

    success = (non_nulls == 0)
    context.result.actual = dict(success=success)

    if success:
        return context.result.success()

    return context.result.fail()


def assert_column_unique(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    column_metrics = metrics.get('tables', {}).get(table, {}).get('columns', {}).get(column)
    if not column_metrics:
        # cannot find the column in the metrics
        return context.result.fail_with_metric_not_found_error(context.table, context.column)

    if context.asserts:
        return context.result.fail_with_no_assert_is_required()

    valids = column_metrics.get('valids')
    distinct = column_metrics.get('distinct')

    success = (valids == distinct)
    context.result.actual = None

    if success:
        return context.result.success()

    return context.result.fail()


def assert_column_exist(context: AssertionContext) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    table_metrics = metrics.get('tables', {}).get(table)
    if not table_metrics:
        # cannot find the table in the metrics
        return context.result.fail_with_metric_not_found_error(context.table, None)

    if context.asserts:
        return context.result.fail_with_no_assert_is_required()

    column_metrics = table_metrics.get('columns', {}).get(column)
    if column_metrics:
        context.result.actual = dict(success=True)
        return context.result.success()

    context.result.actual = dict(success=False)
    return context.result.fail()



class MetricEventHandler:
    def handle_run_start(self):
        pass

    def handle_run_progress(self, total: int, completed: int):
        pass

    def handle_run_end(self):
        pass

    def handle_metric_start(self, metric: str):
        pass

    def handle_metric_progress(self, metric: str, total: int, completed: int):
        pass

    def handle_metric_end(self, metric: str):
        pass

    def handle_param_query_start(self, metric: str, param: str):
        pass

    def handle_param_query_end(self, metric: str):
        pass


class DefaultMetricEventHandler(MetricEventHandler):
    metric_total = 0
    metric_completed = 0
    query_param_total = 0
    query_param_completed = 0

    def handle_run_start(self):
        print("Start Metric Querying")

    def handle_run_progress(self, total: int, completed: int):
        self.metric_total = total
        self.metric_completed = completed

    def handle_run_end(self):
        pass

    def handle_metric_start(self, metric: str):
        print(f"[{self.metric_completed + 1}/{self.metric_total}] querying [{metric}]", flush=True)

    def handle_metric_progress(self, metric: str, total: int, completed: int):
        self.query_param_total = total
        self.query_param_completed = completed

    def handle_metric_end(self, metric: str):
        pass

    def handle_param_query_start(self, metric: str, param: str):
        print(
            f"    [{self.query_param_completed + 1}/{self.query_param_total}] querying [{metric}::{param}] ",
            flush=True
        )

    def handle_param_query_end(self, metric: str):
        pass



import webbrowser


def _send_feedback(url):
    try:
        webbrowser.open(url, new=2)
    except webbrowser.Error:
        print(f"The feedback form: {url}")


class Feedback:

    @staticmethod
    def exec():
        url = "https://docs.google.com/forms/d/e/1FAIpQLSe0J8qC78lqMVxSAJFPub6QXx2NcVY8WLvIVEGthOeQcJHxFQ/viewform?usp=pp_url&entry.2024961102=PipeRider+CLI"
        _send_feedback(url)

    @staticmethod
    def suggest_datasource(data_source_name):
        url = "https://docs.google.com/forms/d/e/1FAIpQLSe0J8qC78lqMVxSAJFPub6QXx2NcVY8WLvIVEGthOeQcJHxFQ/viewform?usp=pp_url&entry.2024961102=PipeRider+CLI&entry.326955045=I+suggest+supporting+a+data+source+for+%5Bfill+me%5D"
        _send_feedback(url)
        pass



from rich.console import Console


class Statistics:
    _instance = None
    statistic = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def reset(self):
        self.statistic = {}

    def add_field(self, field: str, num: int):
        if field not in self.statistic:
            self.statistic[field] = 0
        self.statistic[field] += num

    def add_field_one(self, field: str):
        self.add_field(field, 1)

    def display_statistic(self, action, subject):
        console = Console()
        if self.statistic.get('total', 0) == 0:
            console.print(f"No {subject}s to {action}")
            return
        num_skip = sum(self.statistic.values()) - self.statistic['total']
        if self.statistic['total'] - num_skip == 0:
            console.print(f"No {subject}s to {action}")
        elif num_skip > 0:
            subject = subject + 's' if num_skip > 1 else subject
            msg = ", ".join([f"{k}={v}" for k, v in self.statistic.items() if k != 'total' and v != 0])
            console.print(
                f"{action} {self.statistic['total'] - num_skip} {subject}, skip {num_skip} {subject} ({msg})\n")



import os
import shutil
import tempfile
import uuid
from unittest import TestCase, skip

from sqlalchemy import create_engine, MetaData, Table, Column, Integer

from piperider_cli.assertion_engine import AssertionEngine
from piperider_cli.configuration import FileSystem


def prepare_project_structure():
    from piperider_cli.initializer import _generate_piperider_workspace

    with tempfile.NamedTemporaryFile() as tmp:
        # tmp will be removed after exiting context
        pass
    os.makedirs(tmp.name, exist_ok=True)
    os.chdir(tmp.name)
    _generate_piperider_workspace()

    return tmp.name


def generate_random_directory():
    with tempfile.NamedTemporaryFile() as tmp:
        pass
    os.makedirs(tmp.name, exist_ok=True)
    return tmp.name


def build_assertion_engine(project_dir, table, assertions):
    assertion_file = f'{uuid.uuid4().hex}.yml'
    assertion_path = os.path.join(f'{project_dir}/{FileSystem.PIPERIDER_WORKSPACE_NAME}/assertions',
                                  assertion_file)
    if not os.path.exists(os.path.dirname(assertion_path)):
        os.makedirs(os.path.dirname(assertion_path))

    with open(assertion_path, 'w') as fh:
        fh.write(assertions)

    db_engine = create_engine('sqlite://')
    metadata = MetaData()
    Table(table, metadata, Column('misc', Integer))
    metadata.create_all(db_engine)

    engine = AssertionEngine(db_engine, project_dir)
    engine.load_assertions(config_path=None)
    return engine


@skip("deprecated, skipping")
class UserDefinedTestAssertionsTests(TestCase):

    def setUp(self) -> None:
        self.project_dir = prepare_project_structure()

        # reset the plugin path for testing
        FileSystem.PIPERIDER_ASSERTION_PLUGIN_PATH = os.path.join(f'{self.project_dir}',
                                                                       FileSystem.PIPERIDER_WORKSPACE_NAME,
                                                                       'plugins')

        if not os.path.exists(FileSystem.PIPERIDER_ASSERTION_PLUGIN_PATH):
            os.makedirs(FileSystem.PIPERIDER_ASSERTION_PLUGIN_PATH)

        self.current_module_name = '_user_defined_assertion_functions.py'
        self.custom_asertions = os.path.join(os.path.dirname(__file__), self.current_module_name)

    def test_user_defined_test_from_default_plugin_path(self):
        # put the user defined test function
        shutil.copyfile(self.custom_asertions, os.path.join(f'{FileSystem.PIPERIDER_ASSERTION_PLUGIN_PATH}',
                                                            f'{self.current_module_name}'))

        # use the function in the assertion configuration
        assertions = f"""
        foobarbar:
          tests:
          - name: user-defined-test-test
            assert:
              param1: a
              param2: b
        """
        engine = build_assertion_engine(self.project_dir, 'foobarbar', assertions)

        # invoke customized assertion
        results, exceptions = engine.evaluate_all()

        # check no exceptions
        self.assertEqual([], exceptions)

        # check result
        assertion_result = results[0].result
        self.assertEqual(dict(success=True, exceptions=None),
                         dict(success=assertion_result._success, exceptions=assertion_result._exception))
        self.assertEqual(assertion_result.actual, 'I see you')
        self.assertEqual(assertion_result.expected, {'magic_number': 5566})

    def test_user_defined_test_invoked_from_env_PIPERIDER_PLUGINS(self):
        random_dir = generate_random_directory()

        os.environ['PIPERIDER_PLUGINS'] = random_dir
        print("PIPERIDER_PLUGINS => ", os.environ['PIPERIDER_PLUGINS'])
        # copy this file to PIPERIDER_PLUGINS
        shutil.copyfile(self.custom_asertions,
                        os.path.join(random_dir, f'{self.current_module_name}'))

        # use the function in the assertion configuration
        assertions = f"""
        foobarbar:
          tests:
          - name: user-defined-test-test
            assert:
              param1: a
              param2: b
        """
        engine = build_assertion_engine(self.project_dir, 'foobarbar', assertions)

        # invoke customized assertion
        results, exceptions = engine.evaluate_all()
        self.assertEqual([], exceptions)

        assertion_result = results[0].result
        self.assertEqual(dict(success=True, exceptions=None),
                         dict(success=assertion_result._success, exceptions=assertion_result._exception))
        self.assertEqual(assertion_result.actual, 'I see you')
        self.assertEqual(assertion_result.expected, {'magic_number': 5566})



import os

import inquirer
from rich import box
from rich.console import Console
from rich.prompt import Prompt
from rich.syntax import Syntax
from rich.table import Table

from piperider_cli import clone_directory, safe_load_yaml
from piperider_cli.configuration import Configuration, FileSystem
from piperider_cli.datasource import DataSource, FANCY_USER_INPUT
from piperider_cli.datasource.survey import UserSurveyMockDataSource
from piperider_cli.error import PipeRiderConfigError

console = Console()


def _is_piperider_workspace_exist(workspace_path: str) -> bool:
    if not os.path.exists(workspace_path):
        return False
    elif not os.path.exists(os.path.join(workspace_path, 'config.yml')):
        return False

    return True


def _generate_piperider_workspace() -> bool:
    from piperider_cli import data
    init_template_dir = os.path.join(os.path.dirname(data.__file__), 'piperider-init-template')
    working_dir = FileSystem.PIPERIDER_WORKSPACE_PATH

    if not _is_piperider_workspace_exist(working_dir):
        clone_directory(init_template_dir, working_dir)
        # prepare .gitignore file
        os.rename(os.path.join(working_dir, 'gitignore'), os.path.join(working_dir, '.gitignore'))
        return True
    else:
        # Skip if workspace already exists
        return False


def _ask_user_update_credentials(ds: DataSource):
    console.print(f'\nPlease enter the following fields for {ds.type_name}')
    return ds.ask_credential()


def _ask_user_input_datasource(config: Configuration = None):
    if config is None:
        cls, name = DataSource.ask()
        ds: DataSource = cls(name=name)
        config = Configuration([ds])
        if _ask_user_update_credentials(ds):
            _generate_piperider_workspace()
            config.dump(FileSystem.PIPERIDER_CONFIG_PATH)
            config.dump_credentials(FileSystem.PIPERIDER_CREDENTIALS_PATH, after_init_config=True)
    else:
        if len(config.dataSources) == 1:
            ds = config.dataSources[0]
        else:
            ds = config.ask_for_datasource()
        if not ds.credential:
            console.print(
                f'[[bold yellow]Warning[/bold yellow]] No credential found for \'{ds.type_name}\' datasource \'{ds.name}\'')
            if _ask_user_update_credentials(ds):
                config.dump_credentials(FileSystem.PIPERIDER_CREDENTIALS_PATH)

    ds.show_installation_information()
    if isinstance(ds, UserSurveyMockDataSource):
        ds.send_survey()
        return None
    return config


def _inherit_datasource_from_dbt_project(dbt_project_path, dbt_profiles_dir=None, interactive=True):
    config = safe_load_yaml(FileSystem.PIPERIDER_CONFIG_PATH)
    if config and config.get('dataSources'):
        if interactive is True:
            console.print('[[bold yellow]Warning[/bold yellow]] Found existing configuration. Skip initialization.')
        return config

    dbt_config = Configuration.from_dbt_project(dbt_project_path, dbt_profiles_dir)
    _generate_piperider_workspace()
    dbt_config.dump(FileSystem.PIPERIDER_CONFIG_PATH)

    console.print(f'[[bold green] DBT [/bold green]] Use the existing dbt project file: {dbt_project_path}')
    console.print(
        "[[bold green] DBT [/bold green]] "
        "By default, PipeRider will profile the models and metrics with 'piperider' tag\n"
        "        Apply 'piperider' tag to your models or change the tag in '.piperider/config.yml'\n")

    return dbt_config


def _generate_configuration(dbt_project_path=None, dbt_profiles_dir=None, interactive=True):
    """
    :param dbt_project_path:
    :return: Configuration object
    """
    try:
        config = Configuration.instance()
    except PipeRiderConfigError:
        config = None
    except Exception:
        config = None
        console.print('[[bold yellow]Warning[/bold yellow]] Invalid config.yml')

    if dbt_project_path is None:
        # TODO: mark as deprecated in the future
        console.rule('Deprecated', style='bold red')
        console.print(
            'Non-dbt project is deprecated and will be removed in the future. If you have a strong need for non-dbt project, please contact us by "piperider feedback".\n')
        return _ask_user_input_datasource(config=config)

    if config is not None:
        if interactive:
            console.print('[[bold yellow]Warning[/bold yellow]] Found existing configuration. Skip initialization.')
        return config

    return _inherit_datasource_from_dbt_project(dbt_project_path, dbt_profiles_dir, interactive)


class Initializer():
    @staticmethod
    def exec(working_dir=None, dbt_project_path=None, dbt_profiles_dir=None, interactive=True):
        if working_dir is None:
            working_dir = FileSystem.PIPERIDER_WORKSPACE_PATH

        if _is_piperider_workspace_exist(working_dir) and interactive is True:
            console.print('[bold green]Piperider workspace already exist[/bold green] ')

        # get Configuration object from dbt or user created configuration
        _generate_configuration(dbt_project_path, dbt_profiles_dir, interactive)
        configuration = Configuration.instance()
        configuration.activate_report_directory()

        return configuration

    @staticmethod
    def show_config():

        # show config.yml
        with open(FileSystem.PIPERIDER_CONFIG_PATH, 'r') as f:
            console.rule('.piperider/config.yml')
            config = Syntax(f.read(), "yaml", theme="monokai", line_numbers=True)
            console.print(config)
            console.rule('End of .piperider/config.yml')

    @staticmethod
    def list(report_dir=None):
        working_dir = FileSystem.PIPERIDER_WORKSPACE_PATH

        if _is_piperider_workspace_exist(working_dir):
            config = Configuration.instance()
            list_table = Table(show_header=True, show_edge=True, box=box.SIMPLE_HEAVY)
            list_table.add_column("Datasource", style="cyan", no_wrap=True)
            list_table.add_column("Description", style="magenta", no_wrap=True)
            for ds in config.dataSources:
                fields = ds.get_display_description().split(', ')
                colored_fields = []
                for f in fields:
                    key = f.split('=')[0]
                    value = f.split('=')[-1]
                    colored_fields.append(f"[bold][blue]{key}[/blue][default]=[/default][green]{value}[/green][/bold]")

                list_table.add_row(ds.name, ', '.join(colored_fields))

            console.print(list_table)
        else:
            console.print('[bold red]Piperider workspace does not exist[/bold red] ')

    @staticmethod
    def delete(report_dir=None):

        config = Configuration.instance()
        if config.dbt:
            console.print('[bold yellow]You have connected with a dbt project. '
                          'Please add datasource in the dbt profile directly. [/bold yellow]')
            return

        if FANCY_USER_INPUT:
            questions = [
                inquirer.List('datasource',
                              message='Which datasource do you want to delete?',
                              choices=[(ds.name, ds) for ds in
                                       config.dataSources]),
                inquirer.Confirm('confirm',
                                 message='Are you sure?')
            ]
            answers = inquirer.prompt(questions)
        else:
            console.print('[[yellow]?[/yellow]] Which datasource do you want to delete?')
            idx = 0
            for ds in config.dataSources:
                console.print(f'    [[green]{idx}[/green]] {ds.name}')
                idx += 1
            answer = Prompt.ask('Please enter the number of delete datasource:', choices=[str(i) for i in range(idx)])
            confirm = Prompt.ask('Are you sure? [y/n]', choices=['y', 'n'], default='N')
            answers = {'datasource': config.dataSources[int(answer)], 'confirm': confirm == 'y'}

        if answers['confirm'] is False:
            console.print('[bold red]Abort to delete datasource[/bold red]')
            return

        config.delete_datasource(answers['datasource'])
        config.flush_datasource(FileSystem.PIPERIDER_CONFIG_PATH)
        console.rule('Datasource deleted')

    @staticmethod
    def add(report_dir=None):
        config = Configuration.instance()
        if config.dbt:
            console.print('[bold yellow]You have connected with a dbt project. '
                          'Please add datasource in the dbt profile directly. [/bold yellow]')
            return
        console.rule('Add datasource')
        cls, name = DataSource.ask(exist_datasource=[ds.name for ds in config.dataSources])
        ds: DataSource = cls(name=name)
        config.dataSources.append(ds)
        if _ask_user_update_credentials(ds):
            _generate_piperider_workspace()
            config.flush_datasource(FileSystem.PIPERIDER_CONFIG_PATH)
            config.dump_credentials(FileSystem.PIPERIDER_CREDENTIALS_PATH, after_init_config=True)
            console.rule('Datasource added')
        else:
            console.rule('Abort to add datasource', style='red')



from piperider_cli.error import PipeRiderError
from piperider_cli.feedback import Feedback
from . import DataSource
from .field import ListField, TextField


class UserSurveyMockDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'datasource survey', **kwargs)
        self.fields = [
        ]

    # Always return None to prevent it generate the configue files
    def ask_credential(self):
        super().ask_credential()
        return None

    def validate(self):
        return False, []

    def to_database_url(self):
        return ''

    def verify_connector(self):
        return PipeRiderError('',
                              hint="Thanks for your feedback. We will improve Piperider in the future.\nPlease execute command 'piperider init' with supported data source.")

    def _get_display_description(self):
        return ''

    def get_database(self):
        return ''

    def get_schema(self):
        return ''

    def send_survey(self):
        datasource = self.credential.get('datasource_candidates')
        if datasource == 'Other':
            datasource = self.credential.get('other')
        Feedback.suggest_datasource(datasource)
        pass



import ast
import glob
import os
from unittest import TestCase


class DoYouWriteTheDocsTests(TestCase):

    def test_find_all_builtins(self):
        import piperider_cli

        search_root = os.path.dirname(piperider_cli.__file__)
        docs_root = os.path.abspath(os.path.join(search_root, '../docs/assertions'))
        not_founds = []

        from piperider_cli.assertion_engine.types import custom_registry
        assertion_functions = custom_registry.keys()

        for func in assertion_functions:
            # skip this due to docs structure change
            if func == 'assert_column_value':
                continue
            # check docs
            docs_path = os.path.join(docs_root, f'{func}.md')
            if not os.path.exists(docs_path):
                not_founds.append((func, docs_path))

        for func, filename in not_founds:
            print(f'require docs for function [{func}] at {filename}')
        self.assertEqual([], not_founds)



import math
from dataclasses import dataclass
from enum import Enum

from typing import Callable, Dict, Iterable, List, Optional


class ColumnChangeEntry:
    def __init__(self, column_name: str, base_column_data: Dict, target_column_data: Dict):
        self.column_name = column_name
        self.base_view = ColumnChangeView(base_column_data)
        self.target_view = ColumnChangeView(target_column_data)

    @property
    def changed(self) -> bool:
        return self.base_view != self.target_view


class ColumnChangeView:
    def __init__(self, data: Dict):
        self.data = data

    def __eq__(self, other: "ColumnChangeView"):
        if self.data is None:
            return False
        if other.data is None:
            return False

        # Edited -> types, duplicate, missing(nulls)
        if self.data.get("type") != other.data.get("type"):
            return False

        if self.data.get("duplicates") != other.data.get("duplicates"):
            return False

        if self.data.get("nulls") != other.data.get("nulls"):
            return False

        if self.data.get("distincts") != other.data.get("distincts"):
            return False

        return True


class JoinedTables:
    def __init__(self, base_run: Dict, target_run: Dict):
        self._joined_tables = self._join(base_run.get('tables'), target_run.get('tables'))

    def _merge_keys(self, base: List[str], target: List[str]):
        """
        Merge keys from base, target tables. Unlike default union, it preserves the order for column rename, added, removed.

        :param base: keys for base table
        :param target: keys for base table
        :return: merged keys
        """

        result = []
        while base and target:
            if base[0] == target[0]:
                result.append(base[0])
                base.pop(0)
                target.pop(0)
            elif base[0] in target:
                idx = target.index(base[0])
                for i in target[0:idx]:
                    if i not in result:
                        result.append(i)
                result.append(base[0])
                base.pop(0)
                target = target[idx + 1:]
            else:
                result.append(base[0])
                base.pop(0)

        for c in base:
            if c not in result:
                result.append(c)

        for c in target:
            if c not in result:
                result.append(c)

        return result

    def _join(self, base, target):
        """
        Join base and target to a dict which

        keys = (base keys) +  (target keys)
        result[keys] = {base: {...}, target: {...}

        :param base:
        :param target:
        :return:
        """
        if not base:
            base = dict()
        if not target:
            target = dict()

        keys = self._merge_keys(list(base.keys()), list(target.keys()))
        result = dict()
        for key in keys:
            value = dict()
            value["base"] = base.get(key, {})
            value["target"] = target.get(key, {})
            result[key] = value
        return result

    def columns_changed_iterator(self, table_name: str) -> Iterable[ColumnChangeEntry]:
        all_column_keys, b, t = self._create_columns_and_their_metrics(table_name)

        for column_name in all_column_keys:
            elem = ColumnChangeEntry(column_name, b.get(column_name), t.get(column_name))
            if not elem.changed:
                continue
            yield elem

    def _create_columns_and_their_metrics(self, table_name):
        table = self._joined_tables[table_name]
        b = table.get("base", {}).get("columns", {})
        t = table.get("target", {}).get("columns", {})
        all_column_keys = sorted(set(list(b.keys()) + list(t.keys())))
        return all_column_keys, b, t

    def row_counts(self, table_name):
        table = self._joined_tables[table_name]
        b = table.get("base", {}).get("row_count", math.nan)
        t = table.get("target", {}).get("row_count", math.nan)
        return b, t

    def column_counts(self, table_name):
        table = self._joined_tables[table_name]
        b = table.get("base", {}).get("col_count")
        t = table.get("target", {}).get("col_count")
        return b, t

    def table_data_iterator(self):
        for table_name in self._joined_tables.keys():
            table_data = self._joined_tables[table_name]
            fallback = table_data.get('target') if table_data.get('target') else table_data.get('base')
            if fallback:
                yield table_name, fallback.get('ref_id'), table_data.get('base'), table_data.get('target')


class ChangeType(Enum):
    ADDED = "added"
    REMOVED = "removed"
    MODIFIED = "modified"
    IMPLICIT = "implicit"
    IGNORED = "ignored"

    @property
    def icon_url(self) -> str:
        base_url = "https://raw.githubusercontent.com/InfuseAI/piperider/main/images/icons"

        if self.value == 'added':
            return base_url + "/icon-diff-delta-plus%402x.png"
        if self.value == 'removed':
            return base_url + "/icon-diff-delta-minus%402x.png"
        if self.value == 'modified':
            return base_url + "/icon-diff-delta-explicit%402x.png"
        if self.value == 'implicit':
            return base_url + "/icon-diff-delta-implicit%402x.png"
        return ""

    @property
    def icon_image_tag(self) -> str:
        url = self.icon_url
        if url == "":
            return ""
        return f"""<img src="{url}" width="16px">"""

    def display_changes(self, b, t, format_text: str, *, converter: Callable = None, negative_change: bool = False):
        if self != self.MODIFIED:
            raise ValueError("Only modified type has display for changes")
        if negative_change:
            color = "red" if t > b else "green"
        else:
            color = "green" if t > b else "red"
        sign = "↑" if t > b else "↓"
        diff = t - b
        if math.isnan(diff):
            if math.isnan(b):
                return str(t)
            else:
                return str(b)

        if t == b:
            return str(b)

        if converter:
            return format_text % dict(value=converter(t), color=color, sign=sign, diff=diff)
        return format_text % dict(value=t, color=color, sign=sign, diff=diff)


class ResourceType(Enum):
    MODEL = "model"
    METRIC = "metric"

    @classmethod
    def of(cls, resource_type: str):
        if resource_type == cls.MODEL.value:
            return cls.MODEL
        if resource_type == cls.METRIC.value:
            return cls.METRIC
        raise NotImplementedError(f"no such type: {resource_type}")


@dataclass
class ChangeUnit:
    change_type: ChangeType
    resource_type: ResourceType
    unique_id: str

    @property
    def table_name(self) -> str:
        if self.resource_type == ResourceType.MODEL:
            return self.unique_id.split(".")[-1]
        raise ValueError("It is non sense to get table_name for a non-model resource")


class MetricsChangeView:
    def __init__(self, name: str):
        self.name = name
        self.metric_group: str = None
        self.base_data: Optional[List] = None
        self.target_data: Optional[List] = None
        self.change_type = None

    def update_status(self):
        # added, removed, edited, no changes
        if self.base_data is None and self.target_data is not None:
            self.change_type = "added"
            return

        if self.base_data is not None and self.target_data is None:
            self.change_type = "removed"
            return

        if self.base_data == self.target_data:
            self.change_type = "no-changes"
        else:
            self.change_type = "edited"



import argparse
import json
import tempfile
from pathlib import Path
from typing import Dict, List

import agate
import dbt.flags as flags_module
from dbt.adapters.base import BaseAdapter, BaseRelation, Column as BaseColumn
from dbt.config.project import VarProvider
from dbt.config.runtime import RuntimeConfig
from dbt.contracts.connection import QueryComment
from dbt.contracts.graph.manifest import Manifest, WritableManifest
from dbt.contracts.project import PackageConfig, UserConfig
from dbt.contracts.state import PreviousState
from dbt.node_types import NodeType
from dbt.task.list import ListTask
from dbt.tracking import initialize_from_flags

from piperider_cli.dbt import disable_dbt_compile_stats


def dbt_version():
    from dbt import version
    try:
        version_string = ".".join(version.__version__.split(".")[0:2])
        return version_string
    except BaseException:
        return "unknown"


def dbt_version_obj():
    from packaging import version as v
    from dbt import version as dbt_version
    return v.parse(dbt_version.__version__)


def is_v1_3():
    from packaging import version as v
    dbt_v = dbt_version_obj()
    return v.parse('1.3.0') <= dbt_v < v.parse('v1.4.0')


def is_v1_4():
    from packaging import version as v
    dbt_v = dbt_version_obj()
    return v.parse('1.4.0') <= dbt_v < v.parse('v1.5.0')


def is_lt_v1_5():
    from packaging import version as v
    dbt_v = dbt_version_obj()
    return dbt_v < v.parse('v1.5.0')


def is_v1_5():
    from packaging import version as v
    dbt_v = dbt_version_obj()
    return v.parse('1.5.0') <= dbt_v < v.parse('v1.6.0')


def is_ge_v1_4():
    from packaging import version as v
    dbt_v = dbt_version_obj()
    return dbt_v >= v.parse('1.4.0')


def create_temp_dir():
    return tempfile.mkdtemp()


def load_full_manifest(target_path: str):
    from dbt.adapters.factory import register_adapter
    from dbt.parser.manifest import ManifestLoader

    runtime_config = PrepareRuntimeConfig(target_path)
    register_adapter(runtime_config)

    v = dbt_version()
    if v == '1.5':
        return ManifestLoader.get_full_manifest(
            runtime_config, write_perf_info=False
        )
    elif v == '1.4' or v == '1.3':
        return ManifestLoader.get_full_manifest(
            runtime_config
        )
    raise NotImplementedError(f'dbt-core version: {v} is not supported')


def load_manifest(manifest: Dict):
    v = dbt_version()
    if v == '1.3':
        return _load_manifest_version_13(manifest)

    if v == '1.4':
        return _load_manifest_version_14(manifest)

    if v == '1.5':
        return _load_manifest_version_15(manifest)

    raise NotImplementedError(f'dbt-core version: {v} is not supported')


def get_manifest_schema_version(dct: dict) -> int:
    schema_version = dct.get("metadata", {}).get("dbt_schema_version", None)
    if not schema_version:
        raise ValueError("Manifest doesn't have schema version")
    return int(schema_version.split(".")[-2][-1])


def _load_manifest_version_13(data: Dict):
    from dbt.contracts.util import upgrade_manifest_json
    if get_manifest_schema_version(data) <= 6:
        data = upgrade_manifest_json(data)

    return WritableManifest.from_dict(data)  # type: ignore


def _load_manifest_version_14(data: Dict):
    from dbt.contracts.util import upgrade_manifest_json
    if get_manifest_schema_version(data) <= 7:
        data = upgrade_manifest_json(data)

    return WritableManifest.from_dict(data)  # type: ignore


def _load_manifest_version_15(manifest: Dict):
    from dbt.exceptions import IncompatibleSchemaError

    # return WritableManifest.read_and_check_versions(manifest_path)
    data = manifest

    # Check metadata version. There is a class variable 'dbt_schema_version', but
    # that doesn't show up in artifacts, where it only exists in the 'metadata'
    # dictionary.
    if hasattr(WritableManifest, "dbt_schema_version"):
        if "metadata" in data and "dbt_schema_version" in data["metadata"]:
            previous_schema_version = data["metadata"]["dbt_schema_version"]
            # cls.dbt_schema_version is a SchemaVersion object
            if not WritableManifest.is_compatible_version(previous_schema_version):
                raise IncompatibleSchemaError(
                    expected=str(WritableManifest.dbt_schema_version),
                    found=previous_schema_version,
                )

    return WritableManifest.upgrade_schema_version(data)


class _Adapter(BaseAdapter):
    class NobodyCaredConnectionManager:
        def __init__(self, arg0):
            pass

    ConnectionManager = NobodyCaredConnectionManager

    @classmethod
    def date_function(cls) -> str:
        pass

    @classmethod
    def is_cancelable(cls) -> bool:
        pass

    def list_schemas(self, database: str) -> List[str]:
        pass

    def drop_relation(self, relation: BaseRelation) -> None:
        pass

    def truncate_relation(self, relation: BaseRelation) -> None:
        pass

    def rename_relation(
            self, from_relation: BaseRelation, to_relation: BaseRelation
    ) -> None:
        pass

    def get_columns_in_relation(self, relation: BaseRelation) -> List[BaseColumn]:
        pass

    def expand_column_types(self, goal: BaseRelation, current: BaseRelation) -> None:
        pass

    def list_relations_without_caching(
            self, schema_relation: BaseRelation
    ) -> List[BaseRelation]:
        pass

    def create_schema(self, relation: BaseRelation):
        pass

    def drop_schema(self, relation: BaseRelation):
        pass

    @classmethod
    def quote(cls, identifier: str) -> str:
        pass

    @classmethod
    def convert_text_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        pass

    @classmethod
    def convert_number_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        pass

    @classmethod
    def convert_boolean_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        pass

    @classmethod
    def convert_datetime_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        pass

    @classmethod
    def convert_date_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        pass

    @classmethod
    def convert_time_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        pass


def PrepareRuntimeConfig(target_path: str):
    from piperider_cli.configuration import FileSystem
    project_root = FileSystem.WORKING_DIRECTORY
    profiles_dir = FileSystem.DBT_PROFILES_DIR

    def _get_v13_runtime_config(flags):
        setattr(flags, 'project_dir', project_root)
        setattr(flags, "SEND_ANONYMOUS_USAGE_STATS", False)
        initialize_from_flags()
        return ListTask.ConfigType.from_args(flags)

    def _get_v14_runtime_config(flags):
        setattr(flags, 'project_dir', project_root)
        setattr(flags, "SEND_ANONYMOUS_USAGE_STATS", False)
        initialize_from_flags()
        return ListTask.ConfigType.from_args(flags)

    def _get_v15_runtime_config(flags):
        from dbt.config.runtime import load_project, load_profile

        flags_module.set_flags(flags)
        initialize_from_flags(False, project_root)

        profile = load_profile(
            project_root=project_root,
            cli_vars={}
        )
        project = load_project(
            project_root=project_root,
            version_check=True,
            profile=profile,
            cli_vars={
            }
        )

        return RuntimeConfig.from_parts(
            project,
            profile,
            flags
        )

    flags = flags_module.get_flag_obj()
    setattr(flags, 'target_path', target_path)
    setattr(flags, "WRITE_JSON", None)
    setattr(flags, "exclude", None)
    setattr(flags, "output", "selector")
    setattr(flags, "models", None)
    setattr(flags, "INDIRECT_SELECTION", "eager")
    setattr(flags, "WARN_ERROR", False)
    _configure_warn_error_options(flags)
    setattr(flags, "MACRO_DEBUGGING", False)
    setattr(flags, "PROFILES_DIR", profiles_dir)
    setattr(flags, "cls", ListTask)
    setattr(flags, "profile", None)
    setattr(flags, "target", None)

    v = dbt_version()

    if v == '1.5':
        return _get_v15_runtime_config(flags)
    elif v == '1.4':
        return _get_v14_runtime_config(flags)
    elif v == '1.3':
        return _get_v13_runtime_config(flags)

    raise NotImplementedError(f'dbt-core version: {v} is not supported')


def _configure_warn_error_options(flags):
    if is_ge_v1_4():
        from dbt.helper_types import WarnErrorOptions
        setattr(flags, "WARN_ERROR_OPTIONS", WarnErrorOptions([]))


class _RuntimeConfig(RuntimeConfig):
    def __init__(self):
        from piperider_cli.configuration import FileSystem
        data = {
            "profile_name": "piperider",
            "target_name": "piperider",
            "user_config": UserConfig(),
            "threads": 1,
            "credentials": {"database": "piperider-dbt-functions", "schema": "PUBLIC"},
            "profile_env_vars": {},
            "project_name": "jaffle_shop",
            "version": "0.1",
            "project_root": FileSystem.WORKING_DIRECTORY,
            "model_paths": ["models"],
            "macro_paths": ["macros"],
            "seed_paths": ["seeds"],
            "test_paths": ["tests"],
            "analysis_paths": ["analysis"],
            "docs_paths": ["snapshots", "macros", "seeds", "models", "analysis"],
            "asset_paths": [],
            "target_path": "target",
            "snapshot_paths": ["snapshots"],
            "clean_targets": ["target", "dbt_modules", "logs"],
            "log_path": "None",
            "packages_install_path": "dbt_packages",
            "quoting": {"database": True, "schema": True, "identifier": True},
            "models": {
                "jaffle_shop": {
                    "materialized": "table",
                    "staging": {"materialized": "view"},
                }
            },
            "on_run_start": [],
            "on_run_end": [],
            "dispatch": [],
            "seeds": {},
            "snapshots": {},
            "sources": {},
            "tests": {},
            "metrics": {},
            "exposures": {},
            "vars": VarProvider({}),
            "dbt_version": [],
            "packages": {"packages": PackageConfig(packages=[])},
            "manifest_selectors": {},
            "selectors": {},
            "query_comment": QueryComment(),
            "config_version": 2,
            "unrendered": {
                "project_dict": {},
                "packages_dict": None,
                "selectors_dict": None,
            },
            "project_env_vars": {},
            "cli_vars": {},
            "dependencies": None,
        }

        super().__init__(args=None, **data)

    def validate(self):
        # skip validate
        pass


class _DbtListTask(ListTask):
    def __init__(self):
        self.config = _RuntimeConfig()
        self.args = flags_module.get_flag_obj()
        self.previous_state = None

        if is_v1_5() and hasattr(flags_module, 'set_flags'):
            flags_module.set_flags(self.args)

        # The graph compiler tries to make directories when it initialized itself
        # setattr(self.args, "target_path", "/tmp/piperider-list-task/target_path")
        setattr(self.args, "target_path", 'target')
        setattr(
            self.args,
            "packages_install_path",
            "/tmp/piperider-list-task/packages_install_path",
        )

        # Args for ListTask
        setattr(self.args, "WRITE_JSON", None)
        setattr(self.args, "exclude", None)
        setattr(self.args, "output", "selector")
        setattr(self.args, "models", None)
        setattr(self.args, "INDIRECT_SELECTION", "eager")
        setattr(self.args, "WARN_ERROR", False)
        _configure_warn_error_options(self.args)
        self.args.args = argparse.Namespace()
        self.args.args.cls = ListTask

        # All nodes the compiler building
        self.node_results = []

        try:
            from dbt.task.contextvars import cv_project_root
            if self.config:
                cv_project_root.set(self.config.project_root)
        except Exception:
            # cv_project_root start to be defined since dbt-core v1.5.2
            pass

    def compile_manifest(self):
        if self.manifest is None:
            raise BaseException("compile_manifest called before manifest was loaded")

        adapter = _Adapter(self.config)
        compiler = adapter.get_compiler()
        self.graph = compiler.compile(self.manifest)

    def output_results(self, results):
        # overrider super method: dont print messages or logs
        for result in results:
            self.node_results.append(result)
        return self.node_results

    def load_manifest(self):
        adapter = _Adapter(self.args)
        compiler = adapter.get_compiler()
        self.graph = compiler.compile(self.manifest)
        return


class _InMemoryPreviousState(PreviousState):
    def __init__(self, manifest: Manifest):
        self.manifest = manifest


class ResourceSelector:
    def __init__(self):
        self.selected_node_types = []

    def model(self):
        self.selected_node_types.append(NodeType.Model)
        return self

    def test(self):
        self.selected_node_types.append(NodeType.Test)
        return self

    def seed(self):
        self.selected_node_types.append(NodeType.Seed)
        return self

    def source(self):
        self.selected_node_types.append(NodeType.Source)
        return self

    def metric(self):
        self.selected_node_types.append(NodeType.Metric)
        return self

    def build_selected_set(self):
        return set(
            [x for x in self.selected_node_types if x and isinstance(x, NodeType)]
        )


def list_resources_from_manifest(manifest: Manifest, selector: ResourceSelector = None, select: tuple = None,
                                 state: str = None) -> List[str]:
    task = _DbtListTask()
    task.manifest = manifest

    dbt_flags = task.args
    setattr(dbt_flags, "state", Path(state) if state else None)
    setattr(dbt_flags, "models", None)
    setattr(dbt_flags, "project_target_path", create_temp_dir())

    if is_lt_v1_5():
        flags_module.INDIRECT_SELECTION = 'eager'

    setattr(dbt_flags, "output", "selector")
    setattr(dbt_flags, "selector_name", None)
    setattr(dbt_flags, "resource_types", [])
    if selector is not None:
        setattr(dbt_flags, "resource_types", selector.build_selected_set())

    setattr(dbt_flags, "selector", None)
    setattr(dbt_flags, "select", select)
    if state:
        task.set_previous_state()

    with disable_dbt_compile_stats():
        return task.run()


def list_resources_unique_id_from_manifest(manifest: Manifest):
    task = _DbtListTask()
    task.manifest = manifest

    dbt_flags = task.args
    setattr(dbt_flags, "state", None)
    setattr(dbt_flags, "models", None)
    setattr(dbt_flags, "project_target_path", create_temp_dir())

    if is_lt_v1_5():
        flags_module.INDIRECT_SELECTION = 'eager'

    setattr(dbt_flags, "output", "json")
    setattr(dbt_flags, "output_keys", "unique_id,name,resource_type,original_file_path")
    setattr(dbt_flags, "selector_name", None)
    setattr(dbt_flags, "resource_types", [])

    setattr(dbt_flags, "selector", None)
    setattr(dbt_flags, "select", None)
    with disable_dbt_compile_stats():
        output = task.run()
        return [json.loads(x) for x in output]


def compare_models_between_manifests(
        base_manifest: Manifest,
        altered_manifest: Manifest,
        include_downstream: bool = False,
):
    task = _DbtListTask()
    task.manifest = altered_manifest

    if is_lt_v1_5():
        dbt_flags = task.args
        flags_module.INDIRECT_SELECTION = 'eager'
    else:
        dbt_flags = flags_module.get_flags()

    setattr(dbt_flags, "state", None)
    setattr(dbt_flags, "project_target_path", create_temp_dir())
    setattr(dbt_flags, "models", None)
    setattr(dbt_flags, "selector_name", None)
    setattr(dbt_flags, "output", "selector")
    setattr(dbt_flags, "resource_types", {NodeType.Model})

    setattr(dbt_flags, "selector", None)

    task.previous_state = _InMemoryPreviousState(base_manifest)
    if include_downstream:
        setattr(dbt_flags, "select", ("state:modified+",))
    else:
        setattr(dbt_flags, "select", ("state:modified",))

    if is_ge_v1_4():
        from dbt.exceptions import EventCompilationError as DbtCompilationErr
    else:
        from dbt.exceptions import CompilationException as DbtCompilationErr

    try:
        with disable_dbt_compile_stats():
            return task.run()
    except DbtCompilationErr as e:
        if "does not match any nodes" in e.msg:
            return []
        raise e


def list_modified_with_downstream(
        base_manifest: Manifest,
        altered_manifest: Manifest,
):
    task = _DbtListTask()
    task.manifest = altered_manifest

    if is_lt_v1_5():
        dbt_flags = task.args
        flags_module.INDIRECT_SELECTION = 'eager'
    else:
        dbt_flags = flags_module.get_flags()

    setattr(dbt_flags, "state", None)
    setattr(dbt_flags, "project_target_path", create_temp_dir())
    setattr(dbt_flags, "models", None)
    setattr(dbt_flags, "selector_name", None)
    setattr(dbt_flags, "output", "json")
    setattr(dbt_flags, "output_keys", "unique_id,name,resource_type,original_file_path")
    setattr(dbt_flags, "resource_types", {NodeType.Model, NodeType.Metric})
    setattr(dbt_flags, "selector", None)

    task.previous_state = _InMemoryPreviousState(base_manifest)
    setattr(dbt_flags, "select", ("state:modified+",))

    if is_ge_v1_4():
        from dbt.exceptions import EventCompilationError as DbtCompilationErr
    else:
        from dbt.exceptions import CompilationException as DbtCompilationErr

    try:
        with disable_dbt_compile_stats():
            return [json.loads(x) for x in task.run()]
    except DbtCompilationErr as e:
        if "does not match any nodes" in e.msg:
            return []
        raise e


def list_changes_in_unique_id(
        base_manifest: Manifest,
        target_manifest: Manifest, show_modified_only=False) -> List[Dict[str, str]]:
    task = _DbtListTask()
    task.manifest = target_manifest

    if is_lt_v1_5():
        dbt_flags = task.args
        flags_module.INDIRECT_SELECTION = 'eager'
    else:
        dbt_flags = flags_module.get_flags()

    setattr(dbt_flags, "project_target_path", create_temp_dir())
    setattr(dbt_flags, "state", None)
    setattr(dbt_flags, "models", None)
    setattr(dbt_flags, "selector_name", None)
    setattr(dbt_flags, "output", "json")
    setattr(dbt_flags, "output_keys", "unique_id,name")
    setattr(dbt_flags, "resource_types", None)
    setattr(dbt_flags, "selector", None)

    task.previous_state = _InMemoryPreviousState(base_manifest)

    if show_modified_only:
        setattr(dbt_flags, "select", ("state:modified",))
    else:
        setattr(dbt_flags, "select", None)

    if is_ge_v1_4():
        from dbt.exceptions import EventCompilationError as DbtCompilationErr
    else:
        from dbt.exceptions import CompilationException as DbtCompilationErr

    try:
        with disable_dbt_compile_stats():
            outputs = task.run()
            outputs = [json.loads(x) for x in outputs]
            return outputs
    except DbtCompilationErr as e:
        if "does not match any nodes" in e.msg:
            return []
        raise e



import os.path

from rich.console import Console

from piperider_cli.configuration import Configuration
from piperider_cli.error import RecipeConfigException
from piperider_cli.recipes import select_recipe_file, RecipeConfiguration, execute_recipe_configuration
from piperider_cli.recipes.default_recipe_generator import generate_default_recipe, show_recipe_content

console = Console()


class RecipeExecutor:
    @staticmethod
    def exec(recipe_name: str, auto_generate_default_recipe: bool = True, select: tuple = None, modified: bool = False,
             debug=False, base_branch: str = None, target_branch: str = None):
        config = Configuration.instance()
        recipe_path = select_recipe_file(recipe_name)

        if recipe_name and (select or modified is True):
            console.print(
                "[[bold yellow]Warning[/bold yellow]] The recipe will be ignored when --select or --modified is provided."
            )
        if select:
            console.print(
                f"[[bold green]Select[/bold green]] Manually select the dbt nodes to run by '{','.join(select)}'")
        if recipe_path is None or select or modified is True:
            if auto_generate_default_recipe:
                dbt_project_path = None
                if config.dataSources and config.dataSources[0].args.get('dbt'):
                    dbt_project_path = os.path.relpath(config.dataSources[0].args.get('dbt', {}).get('projectDir'))
                # generate a default recipe
                console.rule("Recipe executor: generate recipe")
                options = None
                if select or modified:
                    options = {}
                    options['select'] = select
                    options['modified'] = modified
                    options['base_branch'] = base_branch
                    options['target_branch'] = target_branch
                recipe = generate_default_recipe(overwrite_existing=False,
                                                 dbt_project_path=dbt_project_path,
                                                 options=options)
                if recipe is None:
                    raise RecipeConfigException(
                        message='Default recipe generation failed.',
                        hint='Please provide a recipe file or feedback an issue to us'
                    )
                show_recipe_content(recipe)
            else:
                raise FileNotFoundError(f"Cannot find the recipe '{recipe_name}'")
        else:
            recipe = RecipeConfiguration.load(recipe_path)
        execute_recipe_configuration(recipe, select=select, modified=modified, debug=debug)

        return recipe
